from dataclasses import dataclass
from enum import auto
from typing import Optional, TYPE_CHECKING, Union

from mstrio.utils.enum_helper import AutoName
from mstrio.utils.exceptions import NotSupportedError
from mstrio.utils.helper import Dictable, exception_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.modeling.schema import UserHierarchy
    from mstrio.modeling.schema.attribute import Attribute


class ObjectSubType(AutoName):
    """String literal used to identify the type of a metadata object.
    Some MicroStrategy APIs make a distinction between an object's type
    and subtype because in some cases (e.g. the different kinds of reports)
    it can be useful to see one "report" concept rather than to have to always
    distinguish between different kinds of reports. But there are cases
    (e.g. filters versus custom-groups or users versus user-groups ) where
    users want to see the different subtypes of the same type as fundamentally
    different.
    Used across different modeling modules e.g. attributes, user hierarchies.
    """

    FILTER = auto()
    CUSTOM_GROUP = auto()
    FILTER_PARTITION = auto()
    SEGMENT = auto()
    TEMPLATE = auto()
    REPORT_GRID = auto()
    REPORT_GRAPH = auto()
    REPORT_ENGINE = auto()
    REPORT_TEXT = auto()
    REPORT_DATAMART = auto()
    REPORT_BASE = auto()
    REPORT_GRID_AND_GRAPH = auto()
    REPORT_NON_INTERACTIVE = auto()
    REPORT_CUBE = auto()
    REPORT_INCREMENT_REFRESH = auto()
    REPORT_TRANSACTION = auto()
    REPORT_EMMA_CUBE = auto()
    REPORT_EMMA_CUBE_IRR = auto()
    REPORT_HYPER_CARD = auto()
    METRIC = auto()
    METRIC_SUBTOTAL = auto()
    SYSTEM_SUBTOTAL = auto()
    METRIC_DMX = auto()
    METRIC_TRAINING = auto()
    METRIC_EXTREME = auto()
    METRIC_REFERENCE_LINE = auto()
    METRIC_RELATIONSHIP = auto()
    STYLE = auto()
    AGG_METRIC = auto()
    FOLDER = auto()
    FOLDER_SYSTEM = auto()
    DEVICE = auto()
    PROMPT = auto()
    PROMPT_BOOLEAN = auto()
    PROMPT_LONG = auto()
    PROMPT_STRING = auto()
    PROMPT_DOUBLE = auto()
    PROMPT_DATE = auto()
    PROMPT_OBJECTS = auto()
    PROMPT_ELEMENTS = auto()
    PROMPT_EXPRESSION = auto()
    PROMPT_EXPRESSION_DRAFT = auto()
    PROMPT_DIMTY = auto()
    PROMPT_BIG_DECIMAL = auto()
    FUNCTION = auto()
    FUNCTION_THIRD_PARTY = auto()
    ATTRIBUTE = auto()
    ATTRIBUTE_ROLE = auto()
    ATTRIBUTE_TRANSFORMATION = auto()
    ATTRIBUTE_ABSTRACT = auto()
    ATTRIBUTE_RECURSIVE = auto()
    ATTRIBUTE_DERIVED = auto()
    FACT = auto()
    DIMENSION_SYSTEM = auto()
    DIMENSION_USER = auto()
    DIMENSION_ORDERED = auto()
    DIMENSION_USER_HIERARCHY = auto()
    LOGICAL_TABLE = auto()
    TABLE_PARTITION_MD = auto()
    TABLE_PARTITION_WH = auto()
    DATAMART_REPORT = auto()
    FACT_GROUP = auto()
    SHORTCUT = auto()
    SHORTCUT_WEAK_REF = auto()
    RESOLUTION = auto()
    ATTRIBUTE_FORM_SYSTEM = auto()
    ATTRIBUTE_FORM_NORMAL = auto()
    ATTRIBUTE_FORM_CUSTOM = auto()
    SCHEMA = auto()
    FORMAT = auto()
    CATALOG = auto()
    CATALOG_DEFN = auto()
    COLUMN = auto()
    COLUMN_NORMAL = auto()
    COLUMN_CUSTOM = auto()
    PROPERTY_GROUP = auto()
    PROPERTY_SET = auto()
    DB_ROLE = auto()
    DB_ROLE_IMPORT = auto()
    DB_ROLE_IMPORT_PRIMARY = auto()
    DB_ROLE_O_AUTH = auto()
    DB_ROLE_REMOTE_DATE_SOURCE = auto()
    DB_ROLE_URL_AUTH = auto()
    DB_ROLE_GENERIC_DATA_CONNECTOR = auto()
    DB_ROLE_DATA_IMPORT_SERVER = auto()
    DB_ROLE_CLOUD_ELEMENT = auto()
    DB_LOGIN = auto()
    DB_CONNECTION = auto()
    PROJECT = auto()
    SERVER_DEF = auto()
    USER = auto()
    USER_GROUP = auto()
    TRANSMITTER = auto()
    CONFIGURATION = auto()
    REQUEST = auto()
    SEARCH = auto()
    INDEXED_SEARCH = auto()
    RELATIONSHIP_SEARCH = auto()
    SEARCH_FOLDER = auto()
    SEARCH_FOLDER_CROSS_PROJECT = auto()
    DATAMART = auto()
    FUNCTION_PACKAGE_DEFINITION = auto()
    ROLE = auto()
    ROLE_TRANSFORMATION = auto()
    SECURITY_ROLE = auto()
    LOCALE = auto()
    CONSOLIDATION = auto()
    CONSOLIDATION_DERIVED = auto()
    CONSOLIDATION_ELEMENT = auto()
    SCHEDULE_EVENT = auto()
    SCHEDULE_OBJECT = auto()
    SCHEDULE_TRIGGER = auto()
    LINK = auto()
    PHYSICAL_TABLE = auto()
    DB_TABLE_PMT = auto()
    DB_TABLE_SOURCE = auto()
    DOCUMENT_DEFINITION = auto()
    REPORT_WRITING_DOCUMENT = auto()
    DOCUMENT_THEME = auto()
    DOSSIER = auto()
    DRILL_MAP = auto()
    DBMS = auto()
    MD_SECURITY_FILTER = auto()
    MONITOR_PERFORMANCE = auto()
    MONITOR_JOBS = auto()
    MONITOR_USER_CONNECTIONS = auto()
    MONITOR_DB_CONNECTIONS = auto()
    PROMPT_ANSWER = auto()
    PROMPT_ANSWER_BOOLEAN = auto()
    PROMPT_ANSWER_LONG = auto()
    PROMPT_ANSWER_STRING = auto()
    PROMPT_ANSWER_DOUBLE = auto()
    PROMPT_ANSWER_DATE = auto()
    PROMPT_ANSWER_OBJECTS = auto()
    PROMPT_ANSWER_ELEMENTS = auto()
    PROMPT_ANSWER_EXPRESSION = auto()
    PROMPT_ANSWER_EXPRESSION_DRAFT = auto()
    PROMPT_ANSWER_DIMTY = auto()
    PROMPT_ANSWER_BIG_DECIMAL = auto()
    PROMPT_ANSWER_INT64 = auto()
    PROMPT_ANSWERS = auto()
    GRAPH_STYLE = auto()
    CHANGE_JOURNAL_SEARCH = auto()
    BLOB_UNKNOWN = auto()
    BLOB_OTHER = auto()
    BLOB_IMAGE = auto()
    BLOB_PROJECT_PACKAGE = auto()
    BLOB_EXCEL = auto()
    BLOB_HTML_TEMPLATE = auto()
    DASHBOARD_TEMPLATE = auto()
    FLAG = auto()
    CHANGE_JOURNAL = auto()
    EXTERNAL_SHORTCUT_UNKNOWN = auto()
    EXTERNAL_SHORTCUT_URL = auto()
    EXTERNAL_SHORTCUT_SNAPSHOT = auto()
    EXTERNAL_SHORTCUT_TARGET = auto()
    RECONCILIATION = auto()
    PALETTE_SYSTEM = auto()
    PALETTE_CUSTOM = auto()
    THRESHOLDS = auto()
    SUBSCRIPTION_ADDRESS = auto()
    SUBSCRIPTION_CONTACT = auto()
    SUBSCRIPTION_INSTANCE = auto()


class TableColumnMergeOption(AutoName):
    REUSE_ANY = auto()
    REUSE_COMPATIBLE_DATA_TYPE = auto()
    REUSE_MATCHED_DATA_TYPE = auto()


class TablePrefixOption(AutoName):
    ADD_DEFAULT_PREFIX = auto()
    ADD_NAMESPACE = auto()


class PhysicalTableType(AutoName):
    NORMAL = auto()
    WAREHOUSE_PARTITION = auto()
    SQL = auto()
    RESERVED = auto()


@dataclass(eq=False)
class SchemaObjectReference(Dictable):
    """Information about an object referenced within the specification
    of another object. An object reference typically contains only enough
    fields to uniquely identify the referenced objects.

    Attributes:
        object_id: object's id, a globally unique identifier used to
            distinguish between metadata objects within the same project
        sub_type:  string literal used to identify the type of a metadata object
        name: the name of the metadata object
        is_embedded: if true indicates that the target object of this reference
            is embedded within this object, bool
    """

    _FROM_DICT_MAP = {'sub_type': ObjectSubType}

    sub_type: ObjectSubType
    object_id: Optional[str] = None
    name: Optional[str] = None
    is_embedded: Optional[bool] = None

    def __eq__(self, other):
        if not isinstance(other, SchemaObjectReference):
            return False

        return self.object_id == other.object_id and self.sub_type == other.sub_type

    def __hash__(self):
        return hash(self.object_id)

    @classmethod
    def create_from(
        cls,
        schema_object: Union["Attribute", "UserHierarchy"],
        is_embedded: bool = None,
    ) -> "SchemaObjectReference":
        """Converts a schema object into a schema object reference

        Args:
            schema_object: a schema object
            is_embeded: a boolean indicating whether the schema object
                is embedded or not

        Returns:
            SchemaObjectReference of the given schema object
        """
        reference_body = {
            "object_id": schema_object.id,
            "name": schema_object.name,
            "sub_type": schema_object.sub_type,
            "is_embeded": is_embedded,
        }
        return cls.from_dict(reference_body)

    def to_object(
        self, connection: "Connection"
    ) -> Union["Attribute", "UserHierarchy"]:
        """Converts a schema object reference into a schema object.

        Args:
            connection: a connection object required to fetch
                the schema object
        """
        if self.sub_type == ObjectSubType.ATTRIBUTE:
            from mstrio.modeling.schema.attribute import Attribute

            return Attribute(connection, id=self.object_id)
        else:
            raise NotSupportedError(
                f"{self.sub_type} object sub type is not supported."
            )


class DataType(Dictable):
    """Representation in the object model for a data-type that could be used
    for a SQL column.

    Attributes:
        type: gross data type of an actual or proposed column in a database.
        precision: for relevant data types, the length of the representation
        scale: for relevant data types, the fixed position used
        in the representation
    """

    class Type(AutoName):
        """String literal used to identify the gross data type of an actual
        or proposed column in a database."""

        UNKNOWN = auto()
        RESERVED = auto()
        INTEGER = auto()
        UNSIGNED = auto()
        NUMERIC = auto()
        DECIMAL = auto()
        REAL = auto()
        FLOAT = auto()
        CHAR = auto()
        FIXED_LENGTH_STRING = auto()
        VARIABLE_LENGTH_STRING = auto()
        BINARY = auto()
        VAR_BIN = auto()
        LONGVARBIN = auto()
        DATE = auto()
        TIME = auto()
        TIME_STAMP = auto()
        N_CHAR = auto()
        N_VAR_CHAR = auto()
        SHORT = auto()
        LONG = auto()
        MB_CHAR = auto()
        BOOL = auto()
        PATTERN = auto()
        N_PATTERN = auto()
        CELL_FORMAT_DATA = auto()
        MISSING = auto()
        UTF8_CHAR = auto()
        INT64 = auto()
        GUID = auto()
        DOUBLE_DOUBLE = auto()

    def __init__(self, type: Type, precision: str, scale: str) -> None:
        self.type = type
        self.precision = precision
        self.scale = scale


class FormReference(Dictable):
    """The reference that identifies a form object within the context of a
    given attribute. When writing back an attribute, either id or name is
    needed to identify a form, and if both are provided, id will take
    the higher priority.

    Attributes:
        id: id of the form
        name: name of the form
    """

    def __init__(self, id: str = None, name: str = None) -> None:
        if id is None and name is None:
            exception_handler(
                "Provide either `id` or `name` of a form object.", AttributeError
            )
        self.id = id
        self.name = name


class AttributeDisplays(Dictable):
    """The collections of report displays and browse displays of the attribute.

    Attributes:
        report_displays: list of an AttributeSorts for report displays
        browse_displays: list of an AttributeSorts for browse displays
    """

    _FROM_DICT_MAP = {
        "report_displays": (
            lambda source, connection: [
                FormReference.from_dict(content, connection) for content in source
            ]
        ),
        "browse_displays": (
            lambda source, connection: [
                FormReference.from_dict(content, connection) for content in source
            ]
        ),
    }

    def __init__(
        self, report_displays: list[FormReference], browse_displays: list[FormReference]
    ) -> None:
        self.report_displays = report_displays
        self.browse_displays = browse_displays


class AttributeSort(Dictable):
    """An individual sort element in an AttributeSorts list.

    Attributes:
        form: A form reference
        ascending: whether the sort is in ascending or descending order
    """

    _FROM_DICT_MAP = {"form": FormReference.from_dict}

    def __init__(self, form: FormReference, ascending: bool = False) -> None:
        self.form = form
        self.ascending = ascending


class AttributeSorts(Dictable):
    """The collections of report sorts and browse sorts of the attribute.

    Attributes:
        report_sorts: list of an AttributeSorts for report sorts
        browse_sorts: list of an AttributeSorts for browse sorts
    """

    _FROM_DICT_MAP = {
        "report_sorts": (
            lambda source, connection: [
                AttributeSort.from_dict(content, connection) for content in source
            ]
        ),
        "browse_sorts": (
            lambda source, connection: [
                AttributeSort.from_dict(content, connection) for content in source
            ]
        ),
    }

    def __init__(
        self,
        report_sorts: Optional[list[AttributeSort]] = None,
        browse_sorts: Optional[list[AttributeSort]] = None,
    ) -> None:
        self.report_sorts = report_sorts
        self.browse_sorts = browse_sorts


@dataclass
class TableColumn(Dictable):
    """An object representation of a physical column that might
    appear in some data source. In addition to representing physical
    columns, we also use this object to represent columns that do not
    actually appear in any data source but which the engine should
    create if it needs to make a column to contain data for some higher
    level construct (e.g. a fact, an attribute form etc.)."""

    _FROM_DICT_MAP = {"data_type": DataType, "sub_type": ObjectSubType}
    data_type: DataType
    column_name: Optional[str] = None  # When retrieved as part of a logical tab
    name: Optional[str] = None  # When retrieved from datasources API
    id: Optional[str] = None
    sub_type: Optional[ObjectSubType] = None
    date_created: Optional[str] = None
    date_modified: Optional[str] = None
    version_id: Optional[str] = None
    primary_locale: Optional[str] = None

    @classmethod
    def from_dict(cls, source, connection, to_snake_case=True) -> 'TableColumn':
        source = source.copy()

        if information := source.get('information', None):
            source.update(information)
            source['id'] = source.pop('objectId')

        return super().from_dict(source, connection, to_snake_case)



"""This is the demo script to show how to manage attributes, attribute forms
and expressions. This script will not work without replacing parameters
with real values. Its basic goal is to present what can be done with
this module and to ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.modeling.schema import (
    Attribute,
    AttributeDisplays,
    AttributeForm,
    AttributeSort,
    AttributeSorts,
    DataType,
    FormReference,
    list_attributes,
    ObjectSubType,
    Relationship,
    SchemaManagement,
    SchemaObjectReference,
    SchemaUpdateType
)
from mstrio.modeling.expression import (
    Expression,
    ExpressionFormat,
    Token,
    ColumnReference,
    Constant,
    Operator,
    FactExpression,
    Variant,
    VariantType,
    Function
)
from mstrio.modeling.schema.table import list_logical_tables, LogicalTable

# For every object we want to reference using a SchemaObjectReference we need
# to provide an Object ID for. For the script to work correctly all occurences
# of `'<object_id>'` and others with form `<some_name>` need to be replaced with
# data specific to the object used.

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

conn = get_connection(workstationData, PROJECT_NAME)

# Function list_logical_tables() might be useful to get table information
# needed to create/alter an Attribute.
# Table name and ID will be used in tables[] in ATTRIBUTE_DATA
logical_tables = list_logical_tables(conn)

# Function LogicalTable.list_columns() might be useful to get table columns information
# needed to create/alter an Attribute.
# Column name and ID will be used in expressions[] in ATTRIBUTE_DATA
logical_table = LogicalTable(conn, id='<logical_table_id>')
table_columns = logical_table.list_columns()

# Define a variable which can be later used in a script
FOLDER_ID = $folder_id  # Insert folder ID here

# Example attribute data.
# Parts of this dictionary will be used in the later parts of this demo script
# Many values in this dictionary should be changed to real ones
ATTRIBUTE_DATA = {
    'name': '<attribute_name>',
    'sub_type': ObjectSubType.ATTRIBUTE,
    'destination_folder': FOLDER_ID,
    'forms': [
        AttributeForm.local_create(
            conn,
            name='<attribute_form_name>',
            alias='<alias>',
            category='<category>',
            display_format=AttributeForm.DisplayFormat.DATE,
            data_type=DataType(
                type=DataType.Type.TIME_STAMP,
                precision=8,
                scale=-2147483648,
            ),
            expressions=[
                FactExpression(
                    expression=Expression(
                        tree=ColumnReference(
                            column_name='<column_name>',
                            object_id='<object_id>',
                        ),
                    ),
                    tables=[
                        SchemaObjectReference(
                            name='<table_name>',
                            sub_type=ObjectSubType.LOGICAL_TABLE,
                            object_id='<object_id>',
                        ),
                        SchemaObjectReference(
                            name='<table_name>',
                            sub_type=ObjectSubType.LOGICAL_TABLE,
                            object_id='<object_id>',
                        ),
                    ],
                ),
            ],
            lookup_table=SchemaObjectReference(
                name='<lookup_table_name>',
                sub_type=ObjectSubType.LOGICAL_TABLE,
                object_id='<object_id>',
            ),
        ),
        AttributeForm.local_create(
            conn,
            name='<attribute_form_name>',
            alias='<alias>',
            category='<cost>',
            display_format=AttributeForm.DisplayFormat.NUMBER,
            data_type=DataType(
                type=DataType.Type.FLOAT,
                precision=8,
                scale=-2147483648,
            ),
            expressions=[
                FactExpression(
                    expression=Expression(
                        tree=ColumnReference(
                            column_name='<column_name>',
                            object_id='<object_id>',
                        ),
                    ),
                    tables=[
                        SchemaObjectReference(
                            name='<table_name>',
                            sub_type=ObjectSubType.LOGICAL_TABLE,
                            object_id='<object_id>',
                        )
                    ],
                ),
            ],
            lookup_table=SchemaObjectReference(
                name='<lookup_table_name>',
                sub_type=ObjectSubType.LOGICAL_TABLE,
                object_id='<object_id>',
            ),
        ),
    ],
    'key_form': FormReference(name='<form_name>'),
    'displays': AttributeDisplays(
        report_displays=[FormReference(name='<form_name>')],
        browse_displays=[FormReference(name='<form_name>')],
    ),
    'attribute_lookup_table': SchemaObjectReference(
        name='<attribute_lookup_table_name>',
        sub_type=ObjectSubType.LOGICAL_TABLE,
        object_id='<object_id>',
    ),
}

# Define a variable which can be later used in a script
ATTRIBUTE_NAME = $attribute_name # Insert name of edited attribute here

# Attributes management
# Get list of attributes, with examples of different conditions
# with expressions represented as trees
list_of_all_atrs = list_attributes(connection=conn)
list_of_limited_atrs = list_attributes(connection=conn, limit=10)
list_of_atrs_by_name = list_attributes(connection=conn, name=ATTRIBUTE_NAME)
list_of_atrs_as_trees = list_attributes(conn, show_expression_as=ExpressionFormat.TREE)

# list attributes with expressions represented as tokens
list_of_atrs_as_tokens = list_attributes(conn, show_expression_as=ExpressionFormat.TOKENS)

# Define a variable which can be later used in a script
ATTRIBUTE_ID = $attribute_id  # Insert ID of edited attribute here

# Get specific attribute by id
attr = Attribute(connection=conn, id=ATTRIBUTE_ID)

# Get specific attribute by name
attr = Attribute(connection=conn, name=ATTRIBUTE_NAME)

# Get specific attribute with expression represented as tree (default argument)
attr = Attribute(conn, name=ATTRIBUTE_NAME, show_expression_as=ExpressionFormat.TREE)

# Get an attribute with expression represented as tokens
attr = Attribute(conn, id=ATTRIBUTE_ID, show_expression_as=ExpressionFormat.TOKENS)

# Listing properties
properies = attr.list_properties()

# Create attribute and get it with expression represented as tree.
# Data from constant `ATTRIBUTE_DATA` defined above in the file is used here
attr = Attribute.create(connection=conn, **ATTRIBUTE_DATA)
attr = Attribute.create(
    connection=conn,
    name=ATTRIBUTE_DATA['name'],
    sub_type=ATTRIBUTE_DATA['sub_type'],
    destination_folder=ATTRIBUTE_DATA['destination_folder'],
    forms=ATTRIBUTE_DATA['forms'],
    key_form=ATTRIBUTE_DATA['key_form'],
    displays=ATTRIBUTE_DATA['displays'],
    attribute_lookup_table=ATTRIBUTE_DATA['attribute_lookup_table'],
)
attr = Attribute.create(
    connection=conn, **ATTRIBUTE_DATA, show_expression_as=ExpressionFormat.TREE
)

# Create an attribute and get it with expression represented as tokens
attr = Attribute.create(conn, **ATTRIBUTE_DATA, show_expression_as=ExpressionFormat.TOKENS)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Define variables which can be later used in a script
ATTRIBUTE_NEW_NAME = $attribute_new_name  # Insert new name of edited attribute here
ATTRIBUTE_DESCRIPTION = $attribute_description  # Insert new description of edited attribute here

# Alter attributes
attr.alter(name=ATTRIBUTE_NEW_NAME, description=ATTRIBUTE_DESCRIPTION)

# Get displays and sorts
displays = attr.displays
sorts = attr.sorts

# Alter displays and sorts
attr.alter(
    displays=AttributeDisplays(
        report_displays=[FormReference(name='<form_name>'), FormReference(name='<form_name>')],
        browse_displays=[FormReference(name='<form_name>')],
    )
)
attr.alter(
    sorts=AttributeSorts(
        report_sorts=[
            AttributeSort(FormReference(name='form_name'), ascending=True),
            AttributeSort(FormReference(name='form_name'))
        ],
        browse_sorts=[AttributeSort(FormReference(name='<form_name>'))],
    )
)

# *Relationship management
# Listing relationship candidates
attr_item = Attribute(conn, name=ATTRIBUTE_NAME)
candidates = attr_item.list_relationship_candidates(already_used=False, to_dictionary=True)

# Listing tables
tables = attr_item.list_tables()

TABLE_NAME = $table_name  # Insert table name here


# Select table with given TABLE_NAME from tables
[choosen_table] = [tab for tab in tables if tab.name == TABLE_NAME]

# Get candidates from choosen table
candidate_attrs = candidates[choosen_table.name]

# Add child and parent
relationship_child: SchemaObjectReference = candidate_attrs[0]
relationship_child_1: SchemaObjectReference = candidate_attrs[1]
relationship_parent: SchemaObjectReference = candidate_attrs[2]
relationship_child_table: SchemaObjectReference = choosen_table

attr_item.add_child(
    relationship_child,
    relationship_type=Relationship.RelationshipType.ONE_TO_MANY,
    table=relationship_child_table
)
attr_item.add_child(
    joint_child=[relationship_child, relationship_child_1],
    relationship_type=Relationship.RelationshipType.MANY_TO_MANY,
    table=relationship_child_table
)
attr_item.add_parent(relationship_parent, table=choosen_table)

# Remove child and parent
attr_item.remove_child(relationship_child)
attr_item.remove_child(joint_child=[relationship_child, relationship_child_1])
attr_item.remove_parent(relationship_parent)

# *Attribute forms management
# Add a form to the attribute
attribute_form = AttributeForm.local_create(
    conn,
    name='<attribute_form_name>',
    alias='<alias>',
    category='<category>',
    display_format=AttributeForm.DisplayFormat.NUMBER,
    data_type=DataType(
        type=DataType.Type.FLOAT,
        precision=8,
        scale=-2147483648,
    ),
    expressions=[
        FactExpression(
            expression=Expression(
                tree=ColumnReference(
                    column_name='<column_name>',
                    object_id='<object_id>',
                ),
            ),
            tables=[
                SchemaObjectReference(
                    name='<table_name>',
                    sub_type=ObjectSubType.LOGICAL_TABLE,
                    object_id='<object_id>',
                )
            ],
        ),
    ],
    lookup_table=SchemaObjectReference(
        name='<lookup_table_name>',
        sub_type=ObjectSubType.LOGICAL_TABLE,
        object_id='<object_id>',
    ),
)

attr.add_form(form=attribute_form)

ATTRIBUTE_FORM_NAME = $attribute_form_name  # Insert attribute form name here


# Get form with specified id. This retrieves the form from the local
# Attribute object, not from the server
attr_form = attr.get_form(name=ATTRIBUTE_FORM_NAME)

ATTRIBUTE_FORM_ALTERED_NAME = $attribute_form_altered_name  # Insert altered attribute form name here
ATTRIBUTE_FORM_DESCRIPTION = $attribute_form_description  # Insert altered attribute form description here

# Alter form
attr.alter_form(
    form_id=attr_form.id, name=ATTRIBUTE_FORM_ALTERED_NAME, description=ATTRIBUTE_FORM_DESCRIPTION
)
attr.alter_form(form_id=attr_form.id, name=ATTRIBUTE_FORM_ALTERED_NAME)

# Removing form with specified id
attr.remove_form(form_id=attr_form.id)

# *AttributeForm's expression management
# Add a fact expression to the attribute form

# create a fact expression with expression specified as tokens
fact_expression = FactExpression(
    expression=Expression(
        tokens=[
            Token(
                type=Token.Type.COLUMN_REFERENCE,
                value='<value>',
                target=SchemaObjectReference(
                    object_id='<object_id>',
                    name='<target_name>',
                    sub_type=ObjectSubType.COLUMN,
                ),
                level=Token.Level.RESOLVED,
                state=Token.State.INITIAL,
            ),
            Token(
                type=Token.Type.END_OF_TEXT,
                value='',
                level=Token.Level.RESOLVED,
                state=Token.State.INITIAL,
            ),
        ],
    ),
    tables=[
        SchemaObjectReference(
            name='<table_name>',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='<object_id>',
        ),
        SchemaObjectReference(
            name='<table_name>',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='<object_id>',
        ),
    ],
)

# create a fact expression with expression specified as tree
fact_expression = FactExpression(
    expression=Expression(
        tree=ColumnReference(
            column_name='<column_name>',
            object_id='<object_id>',
        ),
    ),
    tables=[
        SchemaObjectReference(
            name='<table_name>',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='<object_id>',
        ),
    ],
)

# Add new fact expression to a form
form = attr.get_form(name=ATTRIBUTE_FORM_ALTERED_NAME)
attr.add_fact_expression(form_id=form.id, expression=fact_expression)

# Create a new expression locally using tree format
new_expression = Expression(
    tree=Operator(
        function=Function.ADD,
        children=[
            ColumnReference(
                column_name='<column_name>',
                object_id='<object_id>',
            ),
            Constant(variant=Variant(VariantType.INT32, '1')),
        ],
    ),
)

attr_form = attr.get_form(name=ATTRIBUTE_FORM_ALTERED_NAME)

# Get a fact expression from form by using form name
fact_expression, *_ = [f for f in attr_form.expressions if f.expression.text == 'DAY_DATE']

# Alter the fact expression
attr.alter_fact_expression(
    form_id=attr_form.id, fact_expression_id=fact_expression.id, expression=new_expression
)

# Remove the fact expression from the the attribute form
attr_form = attr.get_form(name=ATTRIBUTE_FORM_ALTERED_NAME)
fact_expression, *_ = attr_form.expressions
attr.remove_fact_expression(form_id=form.id, fact_expression_id=fact_expression.id)

# Deleting attributes
attr.delete(force=True)



import os

import pkg_resources
from setuptools import find_packages, setup
from setuptools.extern.packaging import version as packaging_version

from mstrio import (
    __author__,
    __author_email__,
    __description__,
    __license__,
    __title__,
    __version__,
)

MANIFEST_FILE = 'MANIFEST.in'

if 'version.txt' in os.listdir():
    with open('version.txt') as f:
        dist_version = f.read().strip()
else:
    dist_version = __version__  # define the default version

# Patch Version class to preserve original version string
pkg_resources.safe_version = lambda v: v


class NoNormalizeVersion(packaging_version.Version):
    def __init__(self, version):
        self._orig_version = version
        super().__init__(version)

    def __str__(self):
        return self._orig_version


def find_in_file(text, file):
    if file in os.listdir():
        with open(file, 'r') as f:
            manifest = f.read()
        return text in manifest
    else:
        raise ValueError(f'Could not find file {file}')


packaging_version.Version = NoNormalizeVersion
with open('README.md') as f:
    long_description = f.read()


peer_force_version = [  # peer dependency (dep of dep)
    # but we need to force version for SEC
    # backend only (front peers are
    # handled in code ~20 lines below)
    'urllib3>=1.26.0',  # ver 1.25.x is incompatible
    'Jinja2>=3, <4',  # UI related, but also dev-dep of pandas but required for testing
    'certifi>=2022.12.7',  # SEC safe version
]

# Update test_packages in PyExec integration tests if this list of requirements
# is updated (new package is added or one of them is removed)
# https://github.com/mstr-kiai/PyExec/blob/m2021/tests/integration/test_pyexec.py#L195
requirements = [
    'requests>=2.27, <2.29',
    'requests-futures>=1.0.0, <1.1',
    'pandas>=1.1.5, <1.6',
    'numpy>=1.24, <1.25',
    'tqdm>=4.41, <4.70',
    'packaging>=22.0, <23.0',
    'stringcase>=1.2, <1.3',
    'pyhumps>=3.8',
] + peer_force_version

# Add dependencies for connector-jupyter if connector-jupyter folder is added
if find_in_file('graft connector-jupyter', MANIFEST_FILE):
    requirements += [
        # direct dependencies
        'jupyter-contrib-nbextensions>=0.5.1, <0.6',
        'ipywidgets>=8.0.2, <9',
        # peer, UI related, force version for SEC
        'notebook>=6.4.12',
        'jupyter_core>=4.11.2',
        'lxml>=4.9.1',
        'ipython>=8.10.0, <9',
        'nbconvert>=7, <8',
        'mistune>=2.0.4',
        'setuptools>=65.5.1',
    ]

setup(
    name=__title__,
    python_requires='>=3.10',
    version=dist_version,
    license=__license__,
    description=__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/MicroStrategy/mstrio-py',
    author=__author__,
    author_email=__author_email__,
    project_urls={
        'Bug Tracker': 'https://github.com/MicroStrategy/mstrio-py/issues',
        'Documentation': 'http://www2.microstrategy.com/producthelp/Current/mstrio-py/',
        'Source Code': 'https://github.com/MicroStrategy/mstrio-py',
        'Quick Manual': 'https://www2.microstrategy.com/producthelp/current/MSTR-for-Jupyter/Content/mstr_for_jupyter.htm',  # noqa
    },
    install_requires=requirements,
    extras_require={
        'dev': [
            'flake8',
            'mypy',
            'yapf',
            'nose',
            'coverage',
            'pytest',
            'pytest-cov',
            'isort',
            'pre-commit',
            'flaky',
            'python-decouple',
            'black',
            'flake8-black',
            'build',
        ],  # noqa
    },
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: JavaScript',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
)



from concurrent.futures import as_completed
from contextlib import contextmanager
from functools import wraps
from json import dumps
from typing import TYPE_CHECKING, Optional

from mstrio.api.changesets import (
    commit_changeset_changes,
    create_changeset,
    delete_changeset,
)
from mstrio.utils.helper import (
    auto_match_args,
    delete_none_values,
    get_parallel_number,
    response_handler,
)
from mstrio.utils.sessions import FuturesSessionWithRenewal

if TYPE_CHECKING:
    from mstrio.connection import Connection


def unpack_information(func):
    @wraps(func)
    def unpack_information_inner(*args, **kwargs):
        if kwargs.get('body'):
            info = {
                "information": {
                    "objectId": kwargs['body'].pop('id', None),
                    "subType": kwargs['body'].pop('subType', None),
                    "name": kwargs['body'].pop('name', None),
                    "isEmbedded": kwargs['body'].pop('isEmbedded', None),
                    "description": kwargs['body'].pop('description', None),
                    "dateCreated": kwargs['body'].pop('dateCreated', None),
                    "dateModified": kwargs['body'].pop('dateModified', None),
                    "destinationFolderId": kwargs['body'].pop(
                        'destinationFolderId', None
                    ),
                    "versionId": kwargs['body'].pop('versionId', None),
                    "path": kwargs['body'].pop('path', None),
                    "primaryLocale": kwargs['body'].pop('primaryLocale', None),
                }
            }
            info = delete_none_values(info, recursion=True)
            if info['information']:
                kwargs['body'].update(info)

        resp = func(*args, **kwargs)
        response_json = resp.json()

        if response_json.get('information'):
            response_json.update(response_json.pop('information'))
            response_json['id'] = response_json.pop('objectId', None)
        if response_json.get('tables'):
            response_json = unpack_tables(response_json)
        if (
            isinstance(response_json, dict)
            and response_json.get('subType') == 'logical_table'
        ):
            response_json = unpack_table(response_json)
        resp.encoding, resp._content = 'utf-8', dumps(response_json).encode('utf-8')
        return resp

    return unpack_information_inner


def unpack_table(response_json):
    copy = response_json.copy()

    def __update(objects):
        try:
            [obj.update(obj.pop('information', None)) for obj in objects]
            [obj.update({'id': obj.pop('objectId', None)}) for obj in objects]
        except TypeError:
            # NoneType object is not iterable. Because there are different ways
            # to retrieve a table, sometimes it is already unpacked and don't
            # have attributes / facts. This depends on what getter was used.
            pass

    __update(copy.get('attributes'))
    __update(copy.get('facts'))

    return copy


def unpack_tables(response_json):
    copy = response_json.get('tables').copy()
    for table in copy:
        PHYSICAL_TABLE_FIELD_SET = (
            table.get('physicalTable') and len(table.keys()) == 1
        )  # if retrieved with tables.get_tables(...,fields='physicalTable')
        if PHYSICAL_TABLE_FIELD_SET:
            table.update(table.pop('physicalTable'))
        if table.get('information'):
            table.update(table.pop('information'))
            table['id'] = table.pop('objectId', None)
    return copy


@contextmanager
def changeset_manager(connection):
    response = create_changeset(connection, schema_edit=True)
    changeset_id = response.json()['id']
    try:
        yield changeset_id
        commit_changeset_changes(connection=connection, id=changeset_id)
    finally:
        delete_changeset(connection=connection, id=changeset_id)


def async_get(
    async_wrapper: callable,
    connection: 'Connection',
    ids: list[str],
    error_msg: Optional[str] = None,
    **kwargs,
) -> list[dict]:
    """Asynchronously get results of single object GET requests. GET requests
    requires to have future session param to be used with this function. Threads
    number is set automatically.

    Args:
        async_wrapper: callable async REST API wrapper function
        connection: Connection object
        ids: list of objects ids to be retrieved
        error_msg: optional error message
        kwargs: additional async wrapper arguments to be passed

    Returns:
        List of responses as a list of dicts

    Examples:
        >>> async_get(tables.get_table_async, connection, ['112233','223344'])
    """
    project_id = kwargs.get('project_id') or connection.project_id
    kwargs['project_id'] = project_id
    threads = get_parallel_number(len(ids))
    all_objects = []
    with FuturesSessionWithRenewal(
        connection=connection, max_workers=threads
    ) as session:
        # Extract parameters of the api wrapper and set them using kwargs
        param_value_dict = auto_match_args(
            async_wrapper, kwargs, exclude=['connection', 'session', 'error_msg', 'id']
        )
        futures = [
            async_wrapper(
                session=session, connection=connection, id=id, **param_value_dict
            )
            for id in ids
        ]
        for f in as_completed(futures):
            response = f.result()
            if not response.ok:
                response_handler(response, error_msg, throw_error=False)
            all_objects.append(response.json())

    return all_objects



import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from pandas import DataFrame

from mstrio import config
from mstrio.api import objects, security
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.access_and_security.privilege import Privilege
    from mstrio.server.project import Project
    from mstrio.users_and_groups import UserOrGroup

logger = logging.getLogger(__name__)


@method_version_handler('11.2.0000')
def list_security_roles(
    connection: Connection,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: Optional[int] = None,
    **filters,
):
    """Get all Security Roles stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        to_dictionary(bool, optional): if True, return Security Roles as
            list of dicts
        to_dataframe(bool, optional): if True, return  Security Roles as
            pandas DataFrame
        limit(int, optional): maximum number of security roles returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'description', 'subtype', 'date_created', 'date_modified',
            'version', 'acg', 'owner', 'ext_type']

    Returns:
            List of security roles.
    """
    return SecurityRole._list_security_roles(
        connection,
        to_dictionary=to_dictionary,
        to_dataframe=to_dataframe,
        limit=limit,
        **filters,
    )


@class_version_handler('11.2.0000')
class SecurityRole(Entity, DeleteMixin):
    """A security role is a set of privileges that can be assigned to users and
    reused from project to project. Security roles enable you to assign a
    unique set of privileges to users on a per project basis. They are created
    and maintained at the project source level and assigned to users at the
    project level.

    Attributes:
        connection: A MicroStrategy connection object
        id: Security Role ID
        name: Security Role name
        description: Security Role description
        type: Object type
        subtype: Object subtype
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: owner ID and name
        privileges: Security Role privileges per project
        projects: Project members tuple for the security role
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.SECURITY_ROLE
    _PATCH_PATH_TYPES = {"name": str, "description": str}
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'type',
            'subtype',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'privileges',
            'projects',
            'acg',
            'acl',
        ): security.get_security_role
    }
    _API_PATCH: dict = {
        ('abbreviation'): (objects.update_object, 'partial_put'),
        ('name', 'description'): (security.update_security_role, 'patch'),
    }

    def __init__(
        self,
        connection: Connection,
        name: Optional[str] = None,
        id: Optional[str] = None,
    ):
        """Initialize Security Role object by passing name or id.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            name: name of Security Role
            id: ID of Security Role
        """
        # initialize either by ID or name
        if id is None and name is None:
            helper.exception_handler(
                "Please specify either 'name' or 'id' parameter in the constructor.",
                ValueError,
            )

        if id is None:
            security_roles = SecurityRole._list_security_role_ids(
                connection=connection, name=name
            )
            if security_roles:
                id = security_roles[0]
            else:
                helper.exception_handler(
                    f"There is no Security Role associated with the given name: '"
                    f"{name}'",
                    exception_type=ValueError,
                )
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._projects = kwargs.get("projects")
        self._privileges = kwargs.get("privileges")

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        privileges: Union[
            Union["Privilege", int, str], List[Union["Privilege", int, str]]
        ],
        description: str = "",
    ):
        """Create a new Security Role.

        Args:
            connection(object): MicroStrategy connection object returned
                by 'connection.Connection()'.
            name(string): Name of the Security Role
            privileges: List of privileges which will be assigned to this
                security role. Use privilege IDs or Privilege objects.
            description(string, optional): Description of the Security Role

        Returns:
            Newly created Security Role if the HTTP server has successfully
                created the Security Role.
        """
        # get all project level privileges
        from mstrio.access_and_security.privilege import Privilege

        project_level = [
            priv['id']
            for priv in Privilege.list_privileges(
                connection, to_dictionary=True, is_project_level_privilege='True'
            )
        ]

        # validate and filter passed privileges
        privileges = Privilege._validate_privileges(connection, privileges)
        server_level = list({priv['id'] for priv in privileges} - set(project_level))
        privileges = helper.filter_list_of_dicts(privileges, id=project_level)

        body = {"name": name, "description": description, "privileges": privileges}

        response = security.create_security_role(connection, body)
        if response.ok:
            if server_level:
                msg = (
                    f"Privileges {sorted(server_level)} are server-level and will be "
                    "omitted. Only project-level privileges can be granted by this "
                    "method."
                )
                helper.exception_handler(msg, exception_type=Warning)
            return cls(connection=connection, id=response.json()['id'])

    @classmethod
    def _list_security_roles(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        to_dataframe: bool = False,
        limit: Optional[int] = None,
        **filters,
    ) -> Union[List["SecurityRole"], List[Dict[str, Any]], DataFrame]:
        if to_dictionary and to_dataframe:
            helper.exception_handler(
                "Please select either to_dictionary=True or to_dataframe=True, "
                "but not both.",
                ValueError,
            )

        objects = helper.fetch_objects(
            connection=connection,
            api=security.get_security_roles,
            limit=limit,
            filters=filters,
        )
        if to_dictionary:
            return objects
        elif to_dataframe:
            return DataFrame(objects)
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @classmethod
    def _list_security_role_ids(cls, connection: Connection, **filters) -> List[str]:
        sr_dicts = SecurityRole._list_security_roles(
            connection, to_dictionary=True, **dict(filters)
        )
        return [role.get('id') for role in sr_dicts]  # type: ignore

    def alter(self, name: Optional[str] = None, description: Optional[str] = None):
        """Alter Security Role name or/and description.

        Args:
            name: new name of the Security Role
            description: new description of the Security Role
        """
        func = self.alter
        args = helper.get_args_from_func(func)
        defaults = helper.get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        self._alter_properties(**properties)

    def list_members(self, project_name: Optional[str] = None):
        """List all members of the Security Role. Optionally, filter the
        results by Project name.

        Args:
            project_name(str, optional): Project name
        """
        if project_name is not None:
            [filtered_project] = helper.filter_list_of_dicts(
                self.projects, name=project_name
            )
            members = filtered_project['members']
        else:
            members = []
            for project in self.projects:
                for member in project['members']:
                    members.append(member)
        return members

    def grant_to(
        self,
        members: Union["UserOrGroup", List["UserOrGroup"]],
        project: Union["Project", str],
    ) -> None:
        """Assign users/user groups to a Security Role.

        Args:
            members(list): List of objects or IDs of Users or User Groups which
                will be assigned to this Security Role.
            project(Project, str): Project object or name to which
                this removal will apply.
        """
        from mstrio.server.project import Project
        from mstrio.users_and_groups.user import User
        from mstrio.users_and_groups.user_group import UserGroup

        if isinstance(project, Project):
            project_id = project.id
            project_name = project.name
        elif isinstance(project, str):
            project_list = Project._list_projects(
                connection=self.connection, to_dictionary=True, name=project
            )
            if project_list:
                project_id = project_list[0]['id']
                project_name = project_list[0]['name']
            else:
                helper.exception_handler(
                    f"Project name '{project}' does not exist.", ValueError
                )
        else:
            helper.exception_handler(
                "`project` parameter must be of type str or Project.", TypeError
            )

        # create list of objects from strings/objects/lists
        members_list = members if isinstance(members, list) else [members]
        members_list = [
            obj.id if isinstance(obj, (User, UserGroup)) else str(obj)
            for obj in members_list
        ]
        existing_ids = [
            obj['id'] for obj in self.list_members(project_name=project_name)
        ]
        succeeded = list(set(members_list) - set(existing_ids))
        failed = list(set(existing_ids).intersection(set(members_list)))

        value = {"projectId": project_id, "memberIds": members_list}
        self._update_nested_properties(
            objects=value,
            path="members",
            op='add',
        )
        if config.verbose:
            if succeeded:
                logger.info(f"Granted Security Role '{self.name}' to {succeeded}")
            if failed:
                logger.warning(
                    f"Security Role '{self.name}' already has member(s) {failed}"
                )

    def revoke_from(
        self,
        members: Union["UserOrGroup", List["UserOrGroup"]],
        project: Union["Project", str],
    ) -> None:
        """Remove users/user groups from a Security Role.

        Args:
            members(list): List of objects or IDs of Users or User Groups
                which will be removed from this Security Role.
            project(Project, str): Project object or name
                to which this removal will apply.
        """
        from mstrio.server.project import Project
        from mstrio.users_and_groups.user import User
        from mstrio.users_and_groups.user_group import UserGroup

        if isinstance(project, Project):
            project_id = project.id
            project_name = project.name
        elif isinstance(project, str):
            project_list = Project._list_projects(
                connection=self.connection, to_dictionary=True, name=project
            )
            if project_list:
                project_id = project_list[0]['id']
                project_name = project_list[0]['name']
            else:
                helper.exception_handler(f"Project name '{project}' does not exist.")
        else:
            helper.exception_handler(
                "Project parameter must be of type str or Project.", TypeError
            )

        # create list of objects from strings/objects/lists
        members_list = members if isinstance(members, list) else [members]
        members_list = [
            obj.id if isinstance(obj, (User, UserGroup)) else str(obj)
            for obj in members_list
        ]

        existing_ids = [
            obj['id'] for obj in self.list_members(project_name=project_name)
        ]
        succeeded = list(set(members_list).intersection(set(existing_ids)))
        failed = list(set(members_list) - set(succeeded))

        value = {"projectId": project_id, "memberIds": members_list}
        self._update_nested_properties(
            objects=value,
            path="members",
            op='remove',
        )

        if succeeded and config.verbose:
            logger.info(f"Revoked Security Role '{self.name}' from {succeeded}")
        if failed and config.verbose:
            logger.warning(
                f"Security Role '{self.name}' does not have member(s) {failed}"
            )

    def grant_privilege(
        self,
        privilege: Union[
            Union["Privilege", int, str], List[Union["Privilege", int, str]]
        ],
    ) -> None:
        """Grant new project-level privileges to the Security Role.

        Args:
            privilege: list of privilege objects, ids or names
        """
        # get all project level privileges
        from mstrio.access_and_security.privilege import Privilege

        project_level = [
            priv['id']
            for priv in Privilege.list_privileges(
                self.connection, to_dictionary=True, is_project_level_privilege='True'
            )
        ]

        # validate and filter passed privileges
        privileges = Privilege._validate_privileges(self.connection, privilege)
        server_level = list({priv['id'] for priv in privileges} - set(project_level))
        privileges = helper.filter_list_of_dicts(privileges, id=project_level)

        # create lists for print purposes
        privilege_ids = [priv['id'] for priv in privileges]
        existing_ids = [obj['id'] for obj in self.privileges]
        succeeded = list(set(privilege_ids) - set(existing_ids))
        failed = list(set(existing_ids).intersection(set(privilege_ids)))

        if server_level:
            msg = (
                f"Privileges {sorted(server_level)} are server-level and will be "
                "omitted. Only project-level privileges can be granted by this method."
            )
            helper.exception_handler(msg, exception_type=Warning)

        self._update_nested_properties(
            objects=privileges,
            path="privileges",
            op="addElement",
        )
        if succeeded:
            self.fetch()  # fetch the object properties and set object attributes
            if config.verbose:
                logger.info(f"Granted privilege(s) {succeeded} to '{self.name}'")
        if failed and config.verbose:
            logger.warning(
                f"Security Role '{self.name}' already has privilege(s) {failed}"
            )

    def revoke_privilege(
        self,
        privilege: Union[
            Union["Privilege", int, str], List[Union["Privilege", int, str]]
        ],
    ) -> None:
        """Revoke project-level privileges from the Security Role.

        Args:
            privilege: list of privilege objects, ids or names
        """
        # get all project level privileges
        from mstrio.access_and_security.privilege import Privilege

        project_level = [
            priv['id']
            for priv in Privilege.list_privileges(
                self.connection, to_dictionary=True, is_project_level_privilege='True'
            )
        ]

        # validate and filter passed privileges
        privileges = Privilege._validate_privileges(self.connection, privilege)
        server_level = list({priv['id'] for priv in privileges} - set(project_level))
        privileges = helper.filter_list_of_dicts(privileges, id=project_level)

        # create lists for print purposes
        privilege_ids = [priv['id'] for priv in privileges]
        existing_ids = [obj['id'] for obj in self.privileges]
        succeeded = list(set(privilege_ids).intersection(set(existing_ids)))
        failed = list(set(privilege_ids) - set(succeeded))

        if server_level:
            msg = (
                f"Privilege(s) {sorted(server_level)} are server-level and will be "
                "omitted. Only project-level privileges can be granted by this method."
            )
            helper.exception_handler(msg, exception_type=Warning)

        self._update_nested_properties(
            objects=privileges, path="privileges", op="removeElement"
        )
        if succeeded:
            self.fetch()  # fetch the object properties and set object attributes
            if config.verbose:
                logger.info(f"Revoked privilege(s) {succeeded} from '{self.name}'")
        elif failed and config.verbose:
            logger.warning(
                f"Security Role '{self.name}' does not have privilege(s) {failed}"
            )

    def revoke_all_privileges(self, force: bool = False) -> None:
        """Revoke all granted project-level privileges.

        Args:
            force(bool, optional): If true, overrides the prompt.
        """
        user_input = 'N'
        if not force:
            user_input = input(
                "Are you sure you want to revoke all privileges from Security Role "
                f"'{self.name}'? [Y/N]: "
            )
        if force or user_input == 'Y':
            from mstrio.access_and_security.privilege import Privilege

            project_level = [
                priv['id']
                for priv in Privilege.list_privileges(
                    self.connection,
                    to_dictionary=True,
                    is_project_level_privilege='True',
                )
            ]
            existing_ids = [obj['id'] for obj in self.privileges]
            to_revoke = list(set(project_level).intersection(set(existing_ids)))
            if to_revoke:
                self.revoke_privilege(privilege=to_revoke)
            else:
                logger.warning(
                    f"Security Role '{self.name}' does not have any privilege(s)"
                )

    def list_privileges(self, to_dataframe: bool = False) -> Union[dict, DataFrame]:
        """List ALL privileges for Security Role. Optionally return a
        `DataFrame` object.

        Args:
            to_dataframe: If True, return a `DataFrame` object containing
                privileges
        """
        self.fetch()
        priv_dict = {int(v[1]): k[1] for k, v in [x.items() for x in self.privileges]}

        if to_dataframe:
            df = DataFrame.from_dict(priv_dict, orient='index', columns=['Name'])
            df.index.name = 'ID'
            return df
        else:
            return priv_dict

    def _update_nested_properties(self, objects, path: str, op: str) -> None:
        body = {
            "operationList": [{"op": op, "path": f'/{path}', "value": objects}],
        }

        response = security.update_security_role(self.connection, self.id, body)
        response = response.json()
        if type(response) == dict:
            self._set_object_attributes(**response)

    @property
    def projects(self):
        return self._projects

    @property
    def privileges(self):
        return self._privileges



import logging
from typing import Optional, TYPE_CHECKING, Union

from mstrio.api import datasources
from mstrio.api import tables as tables_api
from mstrio.connection import Connection
from mstrio.modeling.schema import ObjectSubType
from mstrio.modeling.schema.helpers import PhysicalTableType, TableColumn
from mstrio.object_management.search_operations import full_search
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import Entity
from mstrio.utils.helper import fetch_objects, get_valid_project_id
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.modeling.schema.table.logical_table import LogicalTable

NO_PROJECT_ERR_MSG = "You must specify or select a project."

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_physical_tables(
    connection: Connection,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    filters: Optional[dict] = None,
    limit: Optional[int] = None,
    to_dictionary: bool = False,
    include_unassigned_tables: bool = False,
) -> Union[list["PhysicalTable"], list[dict]]:
    """List all physical tables in a project.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection (Connection): Object representation of MSTR Connection.
        project_id (Optional[str], optional): ID of a project. Defaults to None.
        project_name (Optional[str], optional): Name of a project. Defaults to
            None.
        limit(int, optional) limit the number of elements returned.
            If `None` (default), all objects are returned.
        to_dictionary(bool, optional): If True returns a list of dictionaries.
            Defaults to False.
        filters (dict, optional): dict that specifies filter expressions by
            which objects will be filtered locally.
        include_unassigned_tables(bool, optional): if True also returns tables
            which are not assigned to any metadata object. Including those may
            significantly extend time required to gather data.
            Defaults to False.
    Returns:
        Union[list["PhysicalTable"], list[dict]]: A list of PhysicalTable
            objects or dictionaries representing physical tables.
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )
    if include_unassigned_tables:
        tables = full_search(
            connection=connection,
            project=project_id,
            object_types=ObjectTypes.DBTABLE,
            limit=limit,
            **(filters or {}),
        )
    else:
        tables = fetch_objects(
            connection=connection,
            api=tables_api.get_tables,
            limit=limit,
            project_id=project_id,
            filters=filters or {},
            dict_unpack_value="tables",
            fields="physicalTable",
        )

        tables = list({item['id']: item for item in tables}.values())

    return (
        tables
        if to_dictionary
        else PhysicalTable.bulk_from_dict(source_list=tables, connection=connection)
    )


@method_version_handler('11.3.0100')
def list_tables_prefixes(
    connection: Connection,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
):
    """Returns the prefixes for the physical tables

    Args:
        connection: Object representation of MSTR Connection
        project_id (Optional[str], optional): ID of a project. Defaults to None
        project_name (Optional[str], optional): Name of a project. Defaults to
            None.

    Returns:
        A dictionary of prefixes for all the physical tables
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )

    physical_tables = list_physical_tables(
        connection, project_id=project_id, to_dictionary=True
    )
    return {table.get("table_prefix") for table in physical_tables}


@method_version_handler('11.3.0100')
def list_namespaces(
    connection: "Connection",
    id: str,
    refresh: Optional[bool] = None,
    limit: Optional[int] = None,
    **filters,
) -> list[dict]:
    """Get list of namespaces. Optionally filter them by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        id (str): Datasource ID
        refresh (bool, optional): Force refresh
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name']

    Examples:
        >>> list_namespaces(connection, id='A23BBC514D336D5B4FCE919FE19661A3')
    """
    namespaces = fetch_objects(
        connection=connection,
        api=datasources.get_datasource_namespaces,
        dict_unpack_value="namespaces",
        limit=limit,
        filters=filters,
        id=id,
        refresh=refresh,
    )

    return namespaces


@class_version_handler('11.3.0100')
class PhysicalTable(Entity):
    """An object representation of a metadata of physical table. A physical
        table describes the metadata of a warehouse table. It contains a set of
        columns with a schema definition. Currently, two types of physical
        tables are supported. One is called a normal table. It uses the table
        name and namespace to fetch schema definitions from the warehouse. The
        other table is a freeform SQL table. It uses a SQL statement to fetch
        schema definitions.

    Attributes:
        information: An object that contains all of the type-neutral information
            stored in the metadata about an object.
        columns: A list of objects representing a physical column that might
            appear in some data source. In addition to representing physical
            columns, we also use this object to represent columns that do not
            actually appear in any data source but which the engine should
            create if it needs to make a column to contain data for some higher
            level construct (e.g. a fact, an attribute form etc.).
        namespace: The namespace in which a table resides
        table_name: The name of a table
        table_prefix: A table prefix of this physical table
        type: A table type
        sql_statement: This field applies only to a free form sql physical table
    """

    _OBJECT_TYPE: ObjectTypes = ObjectTypes.DBTABLE
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        "columns": [TableColumn.from_dict],
        "table_type": PhysicalTableType,
        "owner": User.from_dict,
    }
    _REST_ATTR_MAP = {"object_id": "id"}

    def __init__(
        self,
        connection: Connection,
        id: str,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ):
        """Initializes a PhysicalTable object. You have to provide ID of a
            physical table object. You can get this ID e.g. using
            `mstrio.project_object.schema.tables.list_physical_tables()`. It
            will try to fetch as much information as possible. To start, it will
            fetch all information related to Entity class. If the physical table
            has at least one logical table associated with it (in a project
            mapped to a connection object), it will fetch additional information
            Info: It's necessary to provide an ID of the table. If you only
                provide the name of a table, it will not be initialized

        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Note:
            When `project_id` is `None` and `project_name` is `None`, then
             its value is overwritten by `project_id` from `connection` object.

        Args:
            connection (Connection): Object representation of MSTR Connection.
            id (str): ID of a physical table object.
            name (str, optional) read-only, Used to improve readability of
                the PhysicalTable class string representation
            project_id (str, optional): ID of a project. Defaults to None.
            project_name (str, optional): Name of a project. Defaults to
                None.

        Examples:
            >>> PhysicalTable(connection, id='5DE0F8934CBD82437D6FA1AFF75F6C58')
        """
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=False if project_name else True,
        )

        super().__init__(connection=connection, object_id=id)
        try:
            # If the physical table is included in a project, more info can be
            # fetched.
            physical_tables: list[dict] = list_physical_tables(
                connection, project_id, to_dictionary=True, filters={'id': id}
            )
            table = physical_tables[0]
            table.update({"table_type": table.pop("type", None)})
            self._set_object_attributes(**table)
        except LookupError:
            pass

    def _init_variables(self, **kwargs) -> None:
        if kwargs.get("id"):
            super()._init_variables(**kwargs)
        elif kwargs.get("information"):
            # available when fetched as a part of a logical table
            information = kwargs.pop("information")
            kwargs.update(
                {
                    "id": information.get("object_id"),
                    "table_type": kwargs.pop("type", None),
                    **information,
                }
            )
            super()._init_variables(**kwargs)

            columns = kwargs.get("columns")
            self.columns = (
                TableColumn.bulk_from_dict(
                    source_list=columns, connection=self.connection
                )
                if columns
                else None
            )
            self.namespace = kwargs.get("namespace")
            self.primary_locale = kwargs.get("primary_locale")

            sub_type = kwargs.get("sub_type")
            self._sub_type = ObjectSubType(sub_type) if sub_type else None
            self.table_name = kwargs.get("table_name")
            self.table_prefix = kwargs.get("table_prefix")

            table_type = kwargs.get("table_type")
            self.table_type = PhysicalTableType(table_type) if table_type else None
            self._version = kwargs.get("version_id")

    def list_dependent_logical_tables(self) -> list["LogicalTable"]:
        """Searches and filters a list of logical tables based on provided
           physical `table_id`.

        Returns:
            list[LogicalTable]: A list of `LogicalTable` objects that are
                dependents of the specified physical table.
        """
        from mstrio.modeling.schema.table.logical_table import list_logical_tables

        logical_tables = list_logical_tables(connection=self.connection)

        return [table for table in logical_tables if table.physical_table.id == self.id]

    def delete(self, force: bool = False) -> None:
        """Searches for all logical table dependents based on provided physical
           `table_id`. If they are found (i.e. physical table is included in
           a project) then information about dependents will be shown and prompt
           will be displayed. If prompt is positive, it will delete all
           dependent logical tables which in turn automatically removes the
           physical table from the project. Note, that table will be deleted
           only if all dependent logical tables have no dependents themselves.

        Args:
            connection (Connection): Object representation of MSTR Connection.
            table_id (str): ID of a physical table.
            force: (bool): If True then prompt will be skipped.
        """
        dependent_logical_tables = self.list_dependent_logical_tables()
        if dependent_logical_tables:
            logger.warning("Following dependent logical tables will be deleted: ")
            logger.warning(",".join(str(table) for table in dependent_logical_tables))
            confirmed_delete = (
                force or input("Do you want to continue? [Y/n]: ").lower() == "y"
            )
            if confirmed_delete:
                [table.delete(force=True) for table in dependent_logical_tables]
                logger.info(f"Successfully removed Warehouse Table with id: {self.id}")
        else:
            logger.error("This table is not included in a project.")



"""This is the demo script to show how to manage tables.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.datasources.datasource_instance import list_connected_datasource_instances
from mstrio.modeling import DataType, SchemaManagement, SchemaUpdateType
from mstrio.modeling.schema.helpers import (
    ObjectSubType, PhysicalTableType, SchemaObjectReference, TableColumn, TableColumnMergeOption
)
from mstrio.modeling.schema.table import (
    list_logical_tables,
    list_physical_tables,
    list_tables_prefixes,
    list_warehouse_tables,
    LogicalTable,
    PhysicalTable,
)

# Define variables which can be later used in a script
PROJECT_ID = $project_id  # Project ID to connect to
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get a list of table prefixes
table_prefixes = list_tables_prefixes(connection=conn)
print(table_prefixes)

# Define a variable which can be later used in a script
PHYSICAL_TABLE_PREFIX = $physical_table_prefix  # Insert prefix of the Table here

# Get a list of physical tables
physical_tables = list_physical_tables(connection=conn, project_id=PROJECT_ID)
print(physical_tables)

# Get a list of logical tables
logical_tables = list_logical_tables(connection=conn)
print(logical_tables)

# List connected datasource instances
connected_datasource_instances = list_connected_datasource_instances(connection=conn)
print(connected_datasource_instances)

# Define variables which can be later used in a script
DATASOURCE_ID = $datasource_id
DATASOURCE_NAME = $datasource_name

# List of all available Warehouse Tables
all_warehouse_tables = list_warehouse_tables(connection=conn)
print(all_warehouse_tables)

# Define a variable which can be later used in a script
WAREHOUSE_TABLE_NAME = $warehouse_table_name  # Insert name of the Table here

# List all Warehouse tables with the name WAREHOUSE_TABLE_NAME
lu_item_tables = list_warehouse_tables(connection=conn, name=WAREHOUSE_TABLE_NAME)
print(lu_item_tables)

# List all Warehouse tables from the DATASOURCE_ID datasource
datasource_tables = list_warehouse_tables(connection=conn, datasource_id=DATASOURCE_ID)
print(datasource_tables)

# Get a Warehouse tables with the name WAREHOUSE_TABLE_NAME
# from the specified DATASOURCE_ID
lu_item_table = list_warehouse_tables(
    connection=conn, name=WAREHOUSE_TABLE_NAME, datasource_id=DATASOURCE_ID
)[0]

print(lu_item_table)

# List Warehouse table columns
print(lu_item_table.list_columns())

# List dependent logical tables of a Warehouse Table.
print(lu_item_table.list_dependent_logical_tables())

# Add a Warehouse Table to a project. This will automatically create both a
# logical table object and physical table object on the server

# Define a variable which can be later used in a script
NEW_LOGICAL_TABLE_NAME = $new_logical_table_name  # Insert name of the logical table

logical_table = lu_item_table.add_to_project(logical_table_name=NEW_LOGICAL_TABLE_NAME)
print(logical_table)
print(logical_table.physical_table)  # List a physical table for a logical table defined above

# List logical tables
logical_tables = list_logical_tables(conn)
print(logical_tables)

# Define variables which can be later used in a script
LOGICAL_TABLE_ID = $logical_table_id  # Insert ID of the logical table
LOGICAL_TABLE_NAME = $logical_table_name  # Insert name of the logical table

# These tables can then be retrieved from the server on demand.
# Get logical table by id
logical_table = LogicalTable(connection=conn, id=LOGICAL_TABLE_ID)
print(logical_table)

# Get logical table by name
logical_table = LogicalTable(connection=conn, name=LOGICAL_TABLE_NAME)
print(logical_table)

# List columns of logical table
print(logical_table.list_columns())

# List physical tables in a project
list_physical_tables(conn, project_id=PROJECT_ID)

# Define a variable which can be later used in a script
PHYSICAL_TABLE_ID = $physical_table_id

# Physical table can be retrieved by its id
physical_table = PhysicalTable(connection=conn, id=PHYSICAL_TABLE_ID)
print(physical_table)

# List dependent logical tables of a physical table
print(physical_table.list_dependent_logical_tables())

# Create logical table alias for physical table
LogicalTable.create_alias(connection=conn, id=PHYSICAL_TABLE_ID)

# Define a variable which can be later used in a script
# The variable below is needed when altering or creating a logical table
LOGICAL_TABLE_DESCRIPTION = $logical_table_description  # Insert description of the logical table

# Create logical table of type normal for a physical table
logical_table = LogicalTable.create(
    connection=conn,
    physical_table=PhysicalTable(connection=conn, id=PHYSICAL_TABLE_ID),
    primary_data_source=SchemaObjectReference(
        object_id=DATASOURCE_ID, sub_type=ObjectSubType.DB_ROLE
    ),
    table_name=LOGICAL_TABLE_NAME,
    table_description=LOGICAL_TABLE_DESCRIPTION,
    sub_type=ObjectSubType.LOGICAL_TABLE,
)
print(logical_table)

# Create logical table of type SQL
logical_table = LogicalTable.create(
    connection=conn,
    physical_table_type=PhysicalTableType.SQL,
    columns=[
        TableColumn(
            data_type=DataType(type=DataType.Type.INTEGER, scale="-2147483648", precision="2"),
            column_name="item_id",
        )
    ],
    sql_statement="SELECT item_id FROM inventory_orders",
    primary_data_source=SchemaObjectReference(
        object_id=DATASOURCE_ID, sub_type=ObjectSubType.DB_ROLE
    ),
    table_name=LOGICAL_TABLE_NAME,
    table_description=LOGICAL_TABLE_DESCRIPTION,
)

# Remove all logical tables added for a specific Warehouse Table from a project.
# All dependent tables will be listed before deletion.
lu_item_table.delete_from_project(force=True)

# List all logical tables based on physical tables of type normal
normal_tables = list_logical_tables(connection=conn, table_type=PhysicalTableType.NORMAL)
print(normal_tables)

# Define a variable which can be later used in a script
NORMAL_TABLE_NAME = $normal_table_name

# Get a table by the name
normal_table = LogicalTable(conn, name=NORMAL_TABLE_NAME)
print(normal_table)

# List all logical tables based on physical tables of type SQL
sql_tables = list_logical_tables(connection=conn, table_type=PhysicalTableType.SQL)
print(sql_tables)

# Define a variable which can be later used in a script
SQL_TABLE_NAME = $sql_table_name

# Get a table by the name
sql_table = LogicalTable(conn, name=SQL_TABLE_NAME)
print(sql_table)

# List all logical tables based on physical tables of type warehouse_partition
warehouse_partition_tables = list_logical_tables(
    connection=conn, table_type=PhysicalTableType.WAREHOUSE_PARTITION
)
print(warehouse_partition_tables)

# List properties for the first logical table in a list
print(logical_tables[0].list_properties())

# List properties for the third logical table in a list
print(logical_tables[2].list_properties())

# Remove logical table
logical_tables[0].delete(force=True)

# Alter table

# Define variables which can be later used in a script
# The variable below is needed when altering a logical table
LOGICAL_TABLE_NAME_ALTERED = $logical_table_name_altered  # Insert altered Table name
# Physical Table
PHYSICAL_TABLE_NAME_ALTERED = $physical_table_name_altered  # Insert altered Table name

# Alter logical table based on physical table of type normal
normal_table.alter(name=LOGICAL_TABLE_NAME_ALTERED)  # logical table name
normal_table.alter(is_true_key=True)
normal_table.alter(logical_size=10)
normal_table.alter(is_logical_size_locked=True)
normal_table.alter(physical_table_object_name=PHYSICAL_TABLE_NAME_ALTERED)
normal_table.alter(physical_table_prefix=PHYSICAL_TABLE_PREFIX)

# Alter logical table based on physical table of type SQL
sql_table.alter(name=LOGICAL_TABLE_NAME_ALTERED)  # logical table name
sql_table.alter(is_true_key=True)
sql_table.alter(logical_size=10)
sql_table.alter(is_logical_size_locked=True)
sql_table.alter(enclose_sql_in_parentheses=True)
sql_table.alter(physical_table_object_name=PHYSICAL_TABLE_NAME_ALTERED)

# You cannot alter logical tables based on physical tables of  type warehouse_partition!

# Update structure of physical table of type normal
normal_table.update_physical_table_structure(TableColumnMergeOption.REUSE_ANY)
normal_table.update_physical_table_structure(TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE)
normal_table.update_physical_table_structure(TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE)

# Update structure of physical table of type SQL
sql_table.update_physical_table_structure(TableColumnMergeOption.REUSE_ANY)
sql_table.update_physical_table_structure(TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE)
sql_table.update_physical_table_structure(TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE)

# Delete a physical table.
physical_table.delete(force=True)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=PROJECT_ID)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])



# flake8: noqa
from .object import list_objects, Object
from .predefined_folders import PredefinedFolders

# isort: off
from .folder import (
    Folder,
    get_my_personal_objects_contents,
    get_predefined_folder_contents,
    list_folders,
)

# isort: on
from .search_enums import (
    CertifiedStatus,
    SearchDomain,
    SearchPattern,
    SearchResultsFormat,
)
from .search_operations import (
    CertifiedStatus,
    full_search,
    get_search_results,
    get_search_suggestions,
    quick_search,
    quick_search_from_object,
    SearchDomain,
    SearchObject,
    SearchPattern,
    SearchResultsFormat,
    start_full_search,
    find_objects_with_id,
)
from .shortcut import get_shortcuts, Shortcut, ShortcutInfoFlags



from typing import Optional

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.time_helper import DatetimeFormats, override_datetime_format


@override_datetime_format(
    DatetimeFormats.YMDHMS.value,
    DatetimeFormats.FULLDATETIME.value,
    ('dateCreated', 'dateModified'),
)
@ErrorHandler(err_msg='Error creating Device.')
def create_device(connection: Connection, body: dict, error_msg: Optional[str] = None):
    """Create a new device.

    Args:
        connection: MicroStrategy REST API connection object
        body: Device creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/v2/devices/"
    return connection.post(url=url, json=body)


@override_datetime_format(
    DatetimeFormats.YMDHMS.value,
    DatetimeFormats.FULLDATETIME.value,
    ('dateCreated', 'dateModified'),
)
@ErrorHandler(err_msg='Error getting Device with ID {id}')
def get_device(connection: Connection, id: str, error_msg: Optional[str] = None):
    """Get device by a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the device
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/v2/devices/{id}"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error deleting Device with ID {id}')
def delete_device(connection: Connection, id: str, error_msg: Optional[str] = None):
    """Delete a device.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the device
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/v2/devices/{id}"
    return connection.delete(url=url)


@override_datetime_format(
    DatetimeFormats.YMDHMS.value,
    DatetimeFormats.FULLDATETIME.value,
    ('dateCreated', 'dateModified'),
)
@ErrorHandler(err_msg='Error updating Device with ID {id}')
def update_device(
    connection: Connection, id: str, body: dict, error_msg: Optional[str] = None
):
    """Update a device.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the device
        body: Device update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/v2/devices/{id}"
    return connection.put(url=url, json=body)


@ErrorHandler(err_msg="Error getting Devices.")
@override_datetime_format(
    DatetimeFormats.YMDHMS.value,
    DatetimeFormats.FULLDATETIME.value,
    ('dateCreated', 'dateModified'),
    'devices',
)
def get_devices(
    connection: Connection,
    device_type: Optional[str] = None,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get information for all devices.

    Args:
        connection: MicroStrategy REST API connection object
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        device_type (str, optional): Device type, Supported values are: email,
            file, ftp, printer, ipad, iphone, android, all.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/v2/devices/"
    params = {'fields': fields, 'deviceType': device_type}
    return connection.get(url=url, params=params)



import logging
from typing import Dict

from packaging import version

from mstrio import config
from mstrio.api import administration
from mstrio.connection import Connection
import mstrio.utils.helper as helper
from mstrio.utils.settings.base_settings import BaseSettings
from mstrio.utils.settings.setting_types import SettingValue
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


class ServerSettings(BaseSettings):
    """Object representation of MicroStrategy I-Server Settings.

    Used to fetch, view, modify, update, export to file, import from file and
    validate Server settings.

    The object can be optionally initialized with `connection`, which will
    automatically fetch the current I-Server settings. If not specified,
    settings can be loaded from file using `import_from()` method. Object
    attributes (representing settings) can be modified manually. Lastly, the
    object can be applied to any project using the `update()` method.

    Attributes:
        settings: settings_value
    """

    _TYPE = "allServerSettings"
    _READ_ONLY = ['historyListRunningStatus']
    _CONVERSION_MAP = {
        'pdfMaxMemoryConsumption': 'B',
        'xmlMaxMemoryConsumption': 'B',
        'excelMaxMemoryConsumption': 'B',
        'htmlMaxMemoryConsumption': 'B',
        'workSetMaxMemoryConsumption': 'KB',
        'catalogMaxMemoryConsumption': 'KB',
    }

    def __init__(self, connection: Connection):
        """Initialize `ServerSettings` object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
        """
        # fix conversion map due to changes in REST Layer
        if version.parse(connection.iserver_version) >= version.parse("11.3.0000"):
            self._CONVERSION_MAP.update(
                {
                    'workSetMaxMemoryConsumption': 'B',
                    'catalogMaxMemoryConsumption': 'B',
                }
            )
        super(BaseSettings, self).__setattr__('_connection', connection)
        self._configure_settings()
        self.fetch()

    @method_version_handler('11.3.0000')
    def fetch(self) -> None:
        """Fetch current I-Server settings and update this `ServerSettings`
        object."""
        super().fetch()

    @method_version_handler('11.3.0000')
    def update(self) -> None:
        """Update the current I-Server settings using this `ServerSettings`
        object."""
        set_dict = self._prepare_settings_push()
        response = administration.update_iserver_settings(self._connection, set_dict)
        if response.status_code in [200, 204] and config.verbose:
            logger.info('I-Server settings updated.')
        if response.status_code == 207:
            helper.exception_handler("Some settings could not be updated.", Warning)

    def _fetch(self) -> Dict:
        settings = administration.get_iserver_settings(self._connection).json()
        return self._prepare_settings_fetch(settings)

    @method_version_handler('11.3.0000')
    def _get_config(self):
        if not ServerSettings._CONFIG:
            response = administration.get_iserver_settings_config(self._connection)
            ServerSettings._CONFIG = response.json()
            super()._get_config()

    def __override_settings_config(self, value: SettingValue) -> None:  # NOSONAR
        # config not accurate, needs override DE179361
        if version.parse(self._connection.iserver_version) < version.parse("11.3.0"):
            if value.name in ['maxJobsPerServer', 'maxInteractiveJobsPerServer']:
                value.max_value = 100000
            elif value.name == 'maxUserConnectionPerServer':
                value.max_value = 1000000
            elif value.name == 'cacheCleanUpFrequency':
                value.min_value = 0
                value.options.append({'name': 'No Limit', 'value': -1})
            elif value.name == 'backupPeriodicity':
                value.min_value = 0
                value.max_value = 99999
            elif value.name == 'xmlMaxMemoryConsumption':
                value.options.clear()
            elif value.name == 'hLAutoDeleteMsgCount':
                value.options.append({'name': 'No Limit', 'value': -1})
            elif value.name == 'catalogMaxMemoryConsumption':
                value.min_value = 1
                value.options.append({"name": "Disabled", "value": -1})



from typing import Optional, Union

from tqdm.auto import tqdm


class ProgressBarMixin:
    _progress_bar: tqdm = None

    def _display_progress_bar(
        self,
        desc: str,
        total: Union[int, float],
        unit: Optional[str] = None,
        initial: Union[int, float] = 0,
        bar_format: Optional[str] = None,
        leave: bool = True,
        **kwargs,
    ):
        """Create and display progress bar. Needs to be closed with
        `close_progress_bar`, when updates no longer needed.

        Args:
            desc : str, optional
                Prefix for the progressbar.
            total : int or float, optional
                The number of expected iterations. If unspecified, len(iterable)
                is used if possible. If float("inf") or as a last resort, only
                basic progress statistics are displayed (no ETA,
                no progressbar). If gui is True and this parameter needs
                subsequent updating, specify an initial arbitrary large
                positive number, e.g. 9e9.
            unit : str, optional
                String that will be used to define the unit of each iteration
                [default: it].
            bar_format : str, optional
                Specify a custom bar string formatting. May impact performance.
                [default: '{l_bar}{bar}{r_bar}'], where l_bar='{desc}:
                {percentage:3.0f}%|' and r_bar='| {n_fmt}/{total_fmt}
                [{elapsed}<{remaining}, ' '{rate_fmt}{postfix}]'
                Possible vars: l_bar, bar, r_bar, n, n_fmt, total, total_fmt,
                percentage, elapsed, elapsed_s, ncols, nrows, desc, unit, rate,
                rate_fmt, rate_noinv, rate_noinv_fmt, rate_inv, rate_inv_fmt,
                postfix, unit_divisor, remaining, remaining_s, eta.
                Note that a trailing ": " is automatically removed after {desc}
                if the latter is empty.
            initial : int or float, optional
                The initial counter value. Useful when restarting a progress
                bar [default: 0]. If using float, consider specifying {n:.3f} or
                similar in bar_format, or specifying unit_scale.
        """
        self._progress_bar = tqdm(
            total=total,
            desc=desc,
            unit=unit,
            bar_format=bar_format,
            initial=initial,
            leave=leave,
            **kwargs,
        )

    def _close_progress_bar(self):
        """Close progress bar when no longer needed. Blocks updates with
        `update_progress_bar_if_needed`."""
        self._progress_bar.close()
        self._progress_bar = None

    def _update_progress_bar_if_needed(
        self,
        new_description: Optional[str] = None,
        update_increment: Union[int, float] = 1,
    ):
        """Updates the state of the progress bar.
        Use it after `display_progress_bar`.
        Args:
            new_description(str): New description of the progress bar
            update_increment: value used when updating progress bar"""
        if self._progress_bar is not None:
            if new_description is not None:
                self._progress_bar.set_description_str(new_description)
            self._progress_bar.update(update_increment)

    def _set_progress_bar_value(self, new_val: Union[int, float]):
        """Sets the value of the progress bar and refreshes it.
        Use it after `display_progress_bar`.
        Args:
            new_val(Union[int, float]): New value of the progress bar"""
        self._progress_bar.n = new_val
        self._progress_bar.refresh()



from dataclasses import dataclass
from enum import auto
import logging
from typing import Any, Dict, Optional, Union

from mstrio.connection import Connection
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable

logger = logging.getLogger(__name__)


@dataclass
class FormatProperty(Dictable):
    """Object that specifies a single format type and it's value

    Attributes:
        type (FormatType): specify what format type is to be set
        value (string): specify what value to set the specified format type to

    FormatType Values:
        Number Category - integer between 0 and 9, default is 9:
            0 - Fixed
            1 - Currency
            2 - Date
            3 - Time
            4 - Percent
            5 - Fraction
            6 - Scientific
            7 - Custom
            8 - Special
            9 - General
        Number Decimal Places - integer between 0 and 30, default is 0
        Number Thousand Separator - boolean (true or false), default is true
        Number Currency Symbol - string of 1-4 characters, default is $
        Number Currency Position - integer between 0 and 3, default is 0:
            0 - at the beginning, for example $10
            1 - at the end, for example 10$
            2 - at the beginning with extra space, for example $ 10
            3 - at the end with extra space, for example 10 $
        Number Format - string of maximum of 255 characters, default value is
            General, example value: ###0.00;[RED]-###0.00
        Number Negative Numbers - integer between 1 and 5, default is 3:
            1 - normal negative format, for example -10
            2 - red color represents negative, -10 would be 10 in red color
            3 - parentheses represent negative, -10 would be (10)
            4 - both red color and parentheses represent negative, -10 would be
                (10) in red color
            5 - minus sign and red color represent negative, -10 would be -10
                in red color
        Alignment Horizontal - integer between 1 and 5, default value is 1:
            1 - General
            2 - Left
            3 - Center
            4 - Right
            5 - Fill
        Alignment Vertical - integer between 1 and 3, default value is 1:
            1 - Top
            2 - Center
            3 - Bottom
        Alignment Text Wrap - boolean (true or false), default is false
        Alignment Text Direction - allowed values are -90, 0 and 90
        Padding Left - double between 0.0 and 100.0, default value is 1.0
        Padding Right - double between 0.0 and 100.0, default value is 1.0
        Padding Top - double between 0.0 and 100.0, default value is 1.0
        Padding Bottom - double between 0.0 and 100.0, default value is 1.0
        Font Name - string representing font, default is Arial, if provided
            font name does not exist or is not supported, default is used
        Font Bold - boolean (true or false), default is false
        Font Italic - boolean (true or false), default is false
        Font Size - integer between 1 and 2147483647, recommended usable range
            is between 8 and 72, other font sizes might render format unreadable
        Font Strikeout - boolean (true or false), default is false
        Font Underline - boolean (true or false), default is false
        Font Color - FormatProperty.Color object representing font color in hex,
            RGB and server readable formats
        Font Script - integer between 1 and 2147483647, default value is 0,
            value represents corresponding font script, if the value has no
            corresponding script, the default will be used
        Border Top Style - integer between 0 and 6, default value is 1:
            0 - empty border
            1 - thin border
            2 - hair border
            3 - dashed border
            4 - dotted border
            5 - thick border
            6 - double border
        Border Right Style - the same rules as Top Style
        Border Left Style - the same rules as Top Style
        Border Bottom Style - the same rules as Top Style
        Border Top Color - FormatProperty.Color object representing border top
         color in hex, RGB and server readable formats
        Border Right Color - FormatProperty.Color object representing border
            right color in hex, RGB and server readable formats
        Border Left Color - FormatProperty.Color object representing border
            left color in hex, RGB and server readable formats
        Border Bottom Color - FormatProperty.Color object representing border
            bottom color in hex, RGB and server readable formats
        Background Fill Style - integer between 0 and 3, default value is 0:
            0 - solid, only has fill color
            1 - transparent
            2 - gradient, has fill color and gradient color
            3 - pattern, has fill color and gradient color
        Background Fill Color - FormatProperty.Color object representing
            background fill color in hex, RGB and server readable formats
        Background Pattern Color - FormatProperty.Color object representing
            background fill color in hex, RGB and server readable formats
        Background Pattern Style - integer between 1 and 18, default value is 1
        Background Gradient Color - FormatProperty.Color object representing
            background gradient color in hex, RGB and server readable formats
        Background Gradient Angle - allowed values are -90, 0 and 90
        Gradient X Offset - allowed values are 0, 50 and 100
        Gradient Y Offset - allowed values are 0, 50 and 100
    """

    class FormatType(AutoName):
        NUMBER_CATEGORY = auto()
        NUMBER_DECIMAL_PLACES = auto()
        NUMBER_THOUSAND_SEPARATOR = auto()
        NUMBER_CURRENCY_SYMBOL = auto()
        NUMBER_CURRENCY_POSITION = auto()
        NUMBER_FORMAT = auto()
        NUMBER_NEGATIVE_NUMBERS = auto()
        ALIGNMENT_HORIZONTAL = auto()
        ALIGNMENT_VERTICAL = auto()
        ALIGNMENT_TEXT_WRAP = auto()
        ALIGNMENT_TEXT_DIRECTION = auto()
        PADDING_LEFT = auto()
        PADDING_RIGHT = auto()
        PADDING_TOP = auto()
        PADDING_BOTTOM = auto()
        FONT_NAME = auto()
        FONT_BOLD = auto()
        FONT_ITALIC = auto()
        FONT_SIZE = auto()
        FONT_STRIKEOUT = auto()
        FONT_UNDERLINE = auto()
        FONT_COLOR = auto()
        FONT_SCRIPT = auto()
        BORDER_TOP_STYLE = auto()
        BORDER_LEFT_STYLE = auto()
        BORDER_BOTTOM_STYLE = auto()
        BORDER_RIGHT_STYLE = auto()
        BORDER_TOP_COLOR = auto()
        BORDER_LEFT_COLOR = auto()
        BORDER_BOTTOM_COLOR = auto()
        BORDER_RIGHT_COLOR = auto()
        BACKGROUND_FILL_COLOR = auto()
        BACKGROUND_PATTERN_COLOR = auto()
        BACKGROUND_PATTERN_STYLE = auto()
        BACKGROUND_FILL_STYLE = auto()
        BACKGROUND_GRADIENT_COLOR = auto()
        BACKGROUND_GRADIENT_ANGLE = auto()
        BACKGROUND_GRADIENT_X_OFFSET = auto()
        BACKGROUND_GRADIENT_Y_OFFSET = auto()

    @dataclass
    class Color(Dictable):
        """A class containing information about the color in different formats.
         It is used as a value for `FormatProperties` that type related
         with colors.

        Attributes:
            hex_value(str): color expressed in hex format (e.g. '#ff02ef')
            server_value(str): color translated for server readable format
            red(str): red component of a RGB format, expressed as decimal
            green(str): green component of a RGB format, expressed as decimal
            blue(str): blue component of a RGB format, expressed as decimal"""

        def _rgb_base_converter(
            self, r: str, g: str, b: str, original_base: int, desired_base: int
        ):
            format_spec = {2: '0>8b', 10: '', 16: '0>2x'}
            return (
                format(int(r, original_base), format_spec[desired_base]),
                format(int(g, original_base), format_spec[desired_base]),
                format(int(b, original_base), format_spec[desired_base]),
            )

        def _init_from_hex(self, hex_val: str) -> str:
            r, g, b = self._rgb_base_converter(
                r=hex_val[1:3],
                g=hex_val[3:5],
                b=hex_val[5:7],
                original_base=16,
                desired_base=2,
            )
            self.server_value = str(int(r + g + b, 2))
            self.hex_value = hex_val
            self.red, self.green, self.blue = self._rgb_base_converter(r, g, b, 2, 10)

        def _init_from_rest(self, rest_val: str) -> str:
            bin_color = format(int(rest_val), '0>24b')
            r, g, b = self._rgb_base_converter(
                r=bin_color[0:8],
                g=bin_color[8:16],
                b=bin_color[16:24],
                original_base=2,
                desired_base=16,
            )
            self.server_value = rest_val
            self.hex_value = f'#{r}{g}{b}'
            self.red, self.green, self.blue = self._rgb_base_converter(r, g, b, 16, 10)

        def _init_from_rgb(self, red: str, green: str, blue: str) -> str:
            r, g, b = self._rgb_base_converter(
                r=red, g=green, b=blue, original_base=10, desired_base=2
            )
            self.server_value = str(int(r + g + b, 2))
            r, g, b = self._rgb_base_converter(
                r=red, g=green, b=blue, original_base=10, desired_base=16
            )
            self.hex_value = f'#{r}{g}{b}'
            self.red, self.green, self.blue = red, green, blue

        hex_value: str
        server_value: str
        red: str
        green: str
        blue: str

        def __init__(
            self,
            hex_value: str | None = None,
            red: int | None = None,
            green: int | None = None,
            blue: int | None = None,
            server_value: str | None = None,
        ):
            """Create an object representing color value of a `FormatProperty`.
             It can be created by providing either hex value, server readable
             value or all three RGB components. Two other representation will
             be generated automatically based on the one provided by the user.
             If more than one representation provided, priority goes as follows:
             server_value->hex_value->RGB values.

            Args:
                hex_value(str): color expressed in hex format (e.g. '#ff02ef')
                server_value(str): color translated for server readable format
                red(int): component of a RGB format, expressed as decimal
                green(int): component of a RGB format, expressed as decimal
                blue(int): component of a RGB format, expressed as decimal"""

            if (
                server_value is not None
                and isinstance(server_value, str)
                and server_value.isnumeric()
            ):
                self._init_from_rest(rest_val=server_value)
            elif hex_value is not None and hex_value[0] == '#' and len(hex_value) == 7:
                self._init_from_hex(hex_val=hex_value)
            elif red in range(256) and green in range(256) and blue in range(256):
                self._init_from_rgb(str(red), str(green), str(blue))
            else:
                raise ValueError(
                    'Invalid parameter for Color value of the format property.'
                )

        def __repr__(self):
            return (
                f'hex_value: {self.hex_value}, rgb: ({self.red}, {self.green}, '
                f'{self.blue}), server_value: {self.server_value}'
            )

    _FROM_DICT_MAP = {
        'type': FormatType,
    }
    type: FormatType
    value: Union[str, "FormatProperty.Color"]

    def to_dict(self, camel_case: bool = True) -> dict:
        if isinstance(self.value, FormatProperty.Color):
            return {'type': self.type.value, 'value': self.value.server_value}
        return {'type': self.type.value, 'value': self.value}

    @classmethod
    def from_dict(
        cls,
        source: Dict[str, Any],
        connection: Optional["Connection"] = None,
        to_snake_case: bool = True,
    ) -> "FormatProperty":
        if source.get('type'):
            if 'color' in source.get('type'):
                return FormatProperty(
                    type=FormatProperty.FormatType(source.get('type')),
                    value=FormatProperty.Color(server_value=source.get('value')),
                )
            return FormatProperty(
                type=FormatProperty.FormatType(source.get('type')),
                value=source.get('value'),
            )


@dataclass
class MetricFormat(Dictable):
    """Object that specifies the formatting of the metric's values and headers

    Attributes:
        values (list[FormatProperty]): list of format properties for the values
            of the metric
        header (list[FormatProperty]): list of format properties for the header
            of the metric
    """

    _FROM_DICT_MAP = {
        'values': (
            lambda source, connection: [
                FormatProperty.from_dict(content, connection) for content in source
            ]
        ),
        'header': (
            lambda source, connection: [
                FormatProperty.from_dict(content, connection) for content in source
            ]
        ),
    }

    values: list[FormatProperty]
    header: list[FormatProperty]



import logging
from functools import partial
from typing import Optional

from pandas import DataFrame, concat

from mstrio import config
from mstrio.api import documents, library, objects
from mstrio.api.schedules import get_contents_schedule
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.object_management import Folder, SearchPattern, search_operations
from mstrio.project_objects import OlapCube, SuperCube
from mstrio.server.environment import Environment
from mstrio.types import ObjectSubTypes
from mstrio.users_and_groups import User, UserGroup, UserOrGroup
from mstrio.utils import helper
from mstrio.utils.cache import CacheSource, ContentCacheMixin
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import (
    CopyMixin,
    DeleteMixin,
    Entity,
    MoveMixin,
    ObjectTypes,
    VldbMixin,
)
from mstrio.utils.helper import (
    IServerError,
    filter_params_for_func,
    get_valid_project_id,
    is_document,
)
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


def list_documents(
    connection: Connection,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: Optional[int] = None,
    name: Optional[str] = None,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    **filters,
) -> list["Document"] | list[dict] | DataFrame:
    """Get all Documents available in the project specified within the
    `connection` object.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
        to_dictionary (bool, optional): if True, return Documents as
            list of dicts
        to_dataframe (bool, optional): if True, return Documents as
            pandas DataFrame
        limit (int, optional): limit the number of elements returned.
            If `None` (default), all objects are returned.
        name (str, optional): characters that the document name must contain
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of documents or list of dictionaries or DataFrame object
    """

    return Document._list_all(
        connection,
        to_dictionary=to_dictionary,
        name=name,
        limit=limit,
        to_dataframe=to_dataframe,
        project_id=project_id,
        project_name=project_name,
        **filters,
    )


def list_documents_across_projects(
    connection: Connection,
    name: Optional[str] = None,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> list["Document"] | list[dict] | DataFrame:
    """Get all Documents stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
        name (string, optional): characters that the document name must contain
        to_dictionary (bool, optional): if True, return Documents as
            list of dicts
        to_dataframe (bool, optional): if True, return Documents as
            pandas DataFrame
        limit (int, optional): limit the number of elements returned. If `None`
            (default), all objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of documents or list of dictionaries or DataFrame object
    """
    project_id_before = connection.project_id
    env = Environment(connection)
    projects = env.list_projects()
    output = DataFrame() if to_dataframe else []
    for project in projects:
        try:
            connection.select_project(project_id=project.id)
        except ValueError:
            if config.verbose:
                logger.info(
                    f'Project {project.name} ({project.id}) is skipped '
                    f'because it does not exist or user has no access '
                    f'to it'
                )
            continue

        docs = Document._list_all(
            connection,
            to_dictionary=to_dictionary,
            name=name,
            limit=limit,
            to_dataframe=to_dataframe,
            **filters,
        )
        if to_dataframe:
            output = concat([output, docs], ignore_index=True)
        else:
            output.extend(docs)

    connection.select_project(project_id=project_id_before)
    return output[:limit]


class Document(Entity, VldbMixin, CopyMixin, MoveMixin, DeleteMixin, ContentCacheMixin):
    """Python representation of MicroStrategy Document object

    _CACHE_TYPE is a variable used by ContentCache class for cache filtering
    purposes.
    """

    _OBJECT_TYPE = ObjectTypes.DOCUMENT_DEFINITION
    _CACHE_TYPE = CacheSource.Type.DOCUMENT
    _API_GETTERS = {**Entity._API_GETTERS, 'recipients': library.get_document}
    _API_PATCH = {
        ('name', 'description', 'folder_id'): (objects.update_object, 'partial_put')
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'certified_info': CertifiedInfo.from_dict,
        'recipients': [User.from_dict],
    }

    def __init__(
        self,
        connection: Connection,
        name: Optional[str] = None,
        id: Optional[str] = None,
    ):
        """Initialize Document object by passing name or id.

        Args:
            connection (object): MicroStrategy connection object returned
                by `connection.Connection()`
            name (string, optional): name of Document
            id (string, optional): ID of Document
        """
        if id is None:
            document = super()._find_object_with_name(
                connection=connection,
                name=name,
                listing_function=partial(
                    self._list_all, search_pattern=SearchPattern.EXACTLY
                ),
            )
            id = document['id']
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._instance_id = ""
        self._recipients = kwargs.get('recipients')
        self._project_id = self.connection.project_id
        self._template_info = kwargs.get('templateInfo')
        self._folder_id = None

    def list_properties(self):
        """List properties for the document.

        Returns:
            A list of all document properties."""
        properties = {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'type': self.type,
            'subtype': self.subtype,
            'ext_type': self.ext_type,
            'ancestors': self.ancestors,
            'certified_info': self.certified_info,
            'comments': self.comments,
            'date_created': self.date_created,
            'date_modified': self.date_modified,
            'instance_id': self.instance_id,
            'owner': self.owner,
            'recipients': self.recipients,
            'version': self.version,
            'template_info': self.template_info,
            'acg': self.acg,
            'acl': self.acl,
        }
        return properties

    def alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        folder_id: Optional[Folder | str] = None,
    ):
        """Alter Document name, description and/or folder id.

        Args:
            name (string, optional): new name of the Document
            description (string, optional): new description of the Document
            folder_id (string | Folder, optional): A globally unique identifier
                used to distinguish between metadata objects within the same
                project. It is possible for two metadata objects in different
                projects to have the same Object Id.
        """
        description = description or self.description
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)
        if folder_id:
            self._folder_id = folder_id

    def __validate_user(self, recipient_id: str) -> str | None:
        try:
            User(self.connection, id=recipient_id)
        except IServerError:
            if config.verbose:
                logger.info(f'{recipient_id} is not a valid value for User ID')
            return None
        return recipient_id

    def publish(self, recipients: Optional[UserOrGroup | list[UserOrGroup]] = None):
        """Publish the document for authenticated user. If `recipients`
        parameter is specified publishes the document for the given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup], optional): list of users
                or user groups to publish the document to (can be a list of IDs
                or a list of User and UserGroup elements)
        """
        if not isinstance(recipients, list) and recipients is not None:
            recipients = [recipients]

        if recipients is None:
            recipients = [self.connection.user_id]
        elif all([isinstance(el, User) for el in recipients]):
            recipients = [recipient.id for recipient in recipients]
        elif all([isinstance(el, UserGroup) for el in recipients]):
            users = [user for group in recipients for user in group.members]
            recipients = [user["id"] for user in users]
        elif any([not isinstance(el, str) for el in recipients]):
            raise ValueError(
                'Please provide either list of User, UserGroup or str elements.'
            )
        for recipient in recipients:
            if not self.__validate_user(recipient):
                recipients.remove(recipient)
        body = {'id': self.id, 'recipients': recipients}
        library.publish_document(self.connection, body)
        self.fetch(attr='recipients')

    def unpublish(self, recipients: Optional[UserOrGroup | list[UserOrGroup]] = None):
        """Unpublish the document for all users it was previously published to.
        If `recipients` parameter is specified unpublishes the document for the
        given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup], optional): list of users
                or user groups to publish the document to (can be a list of IDs
                or a list of User and UserGroup elements)
        """

        if recipients is None:
            library.unpublish_document(self.connection, id=self.id)
        else:
            if not isinstance(recipients, list):
                recipients = [recipients]
            if all([isinstance(el, User) for el in recipients]):
                recipients = [recipient.id for recipient in recipients]
            elif all([isinstance(el, UserGroup) for el in recipients]):
                users = [user for group in recipients for user in group.members]
                recipients = [user["id"] for user in users]
            elif any([not isinstance(el, str) for el in recipients]):
                raise ValueError(
                    'Please provide either list User and UserGroup elements or str '
                    'elements.'
                )
            for user_id in recipients:
                if self.__validate_user(user_id):
                    library.unpublish_document_for_user(
                        self.connection, document_id=self.id, user_id=user_id
                    )
        self.fetch(attr='recipients')

    @method_version_handler('11.3.0600')
    def list_available_schedules(
        self, to_dictionary: bool = False
    ) -> list["Schedule"] | list[dict]:
        """Get a list of schedules available for the object instance.

        Args:
            to_dictionary (bool, optional): If True returns a list of
                dictionaries, otherwise returns a list of Schedules.
                False by default.

        Returns:
            List of Schedule objects or list of dictionaries.
        """
        schedules_list_response = (
            get_contents_schedule(
                connection=self.connection,
                project_id=self.connection.project_id,
                body={'id': self.id, 'type': 'document'},
            ).json()
        ).get('schedules')
        if to_dictionary:
            return schedules_list_response
        else:
            return [
                Schedule.from_dict(connection=self.connection, source=schedule_id)
                for schedule_id in schedules_list_response
            ]

    def share_to(self, users: UserOrGroup | list[UserOrGroup]):
        """Shares the document to the listed users' libraries.

        Args:
            users(UserOrGroup | list[UserOrGroup]): list of users or user
                groups to publish the document to (can be a list of IDs or a
                list of User and UserGroup elements).
        """
        self.publish(users)

    @classmethod
    def _list_all(
        cls,
        connection: Connection,
        search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
        name: Optional[str] = None,
        to_dictionary: bool = False,
        to_dataframe: bool = False,
        limit: Optional[int] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        **filters,
    ) -> list["Document"] | list[dict] | DataFrame:
        if to_dictionary and to_dataframe:
            helper.exception_handler(
                "Please select either `to_dictionary=True` or `to_dataframe=True`, but "
                "not both.",
                ValueError,
            )
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=False if project_name else True,
        )

        objects = search_operations.full_search(
            connection,
            object_types=ObjectSubTypes.REPORT_WRITING_DOCUMENT,
            project=project_id,
            name=name,
            pattern=search_pattern,
            **filters,
        )
        documents = [obj for obj in objects if is_document(obj['view_media'])]

        documents = documents[:limit] if limit else documents

        if to_dictionary:
            return documents
        elif to_dataframe:
            return DataFrame(documents)
        else:
            return [
                cls.from_dict(source=document, connection=connection)
                for document in documents
            ]

    def get_connected_cubes(self) -> list[SuperCube | OlapCube]:
        """Lists cubes used by this document.

        Returns:
            A list of cubes used by the document."""
        cubes = documents.get_cubes_used_by_document(self.connection, self.id).json()
        ret_cubes = [helper.choose_cube(self.connection, cube) for cube in cubes]
        return [tmp for tmp in ret_cubes if tmp]  # remove `None` values

    @property
    def instance_id(self):
        if self._instance_id == '':
            body = {"resolveOnly": True, "persistViewState": True}
            response = documents.create_new_document_instance(
                connection=self.connection, document_id=self.id, body=body
            )
            self._instance_id = response.json()['mid']
        return self._instance_id

    @property
    def folder_id(self):
        if not self._folder_id:
            self._folder_id = (
                next(folder['id'] for folder in self.ancestors if folder['level'] == 1)
                if self.ancestors
                else None
            )
        return self._folder_id

    @property
    def recipients(self):
        return self._recipients

    @property
    def template_info(self):
        return self._template_info



from typing import Optional

from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg='Error creating a metric')
def create_metric(
    connection: Connection,
    body: dict,
    show_expression_as: Optional[list[str]] = None,
    show_filter_tokens: bool = False,
):
    """Create a new metric in the changeset,
    based on the definition provided in request body.

    Args:
        connection: MicroStrategy REST API connection object
        body: Metric creation data
        show_expression_as: Specifies the format in which the expressions are
           returned in response
           Available values: "tokens", "tree"
           If omitted, the expression is returned in "text" format
           If "tree", the expression is returned in "text" and "tree" formats.
           If "tokens", the expression is returned in "text" and "tokens"
           formats.
        show_filter_tokens: Specifies whether the "condition" in threshold is
            returned in "tokens" format, along with "text" and "tree" formats.
            If omitted or false, only "text" and "tree" formats are returned.
            If true, all "text", "tree" and "tokens" formats are returned.
    Return:
        HTTP response object. Expected status: 201
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            url=f"{connection.base_url}/api/model/metrics",
            headers={"X-MSTR-MS-Changeset": changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showFilterTokens': str(show_filter_tokens).lower(),
            },
            json=body,
        )


@unpack_information
@ErrorHandler(err_msg='Error getting metric with ID: {id}')
def get_metric(
    connection: Connection,
    id: str,
    changeset_id: Optional[str] = None,
    show_expression_as: Optional[list[str]] = None,
    show_filter_tokens: bool = False,
):
    """Get definition of a single metric by id

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a metric
        changeset_id: ID of a changeset
        show_expression_as: Specifies the format in which the expressions
           are returned in response
           Available values: "tokens", "tree"
           If omitted, the expression is returned in "text" format
           If "tree", the expression is returned in "text" and "tree" formats.
           If "tokens", the expression is returned in "text" and "tokens"
           formats.
        show_filter_tokens: Specifies whether the "condition" in threshold is
            returned in "tokens" format, along with "text" and "tree" formats.
            If omitted or false, only "text" and "tree" formats are returned.
            If true, all "text", "tree" and "tokens" formats are returned.

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.get(
        url=f'{connection.base_url}/api/model/metrics/{id}',
        headers={'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showFilterTokens': str(show_filter_tokens).lower(),
        },
    )


@unpack_information
@ErrorHandler(err_msg='Error updating metric with ID: {id}')
def update_metric(
    connection: Connection,
    id: str,
    body: dict,
    show_expression_as: Optional[list[str]] = None,
    show_filter_tokens: bool = False,
):
    """Update a specific metric in the changeset,
    based on the definition provided in request body.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a metric
        body: Metric creation data
        show_expression_as: Specifies the format in which the expressions are
           returned in response
           Available values: "tokens", "tree"
           If omitted, the expression is returned in "text" format
           If "tree", the expression is returned in "text" and "tree" formats.
           If "tokens", the expression is returned in "text" and "tokens"
           formats.
        show_filter_tokens: Specifies whether the "condition" in threshold is
            returned in "tokens" format, along with "text" and "tree" formats.
            If omitted or false, only "text" and "tree" formats are returned.
            If true, all "text", "tree" and "tokens" formats are returned.

    Return:
        HTTP response object. Expected status: 201
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            url=f'{connection.base_url}/api/model/metrics/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showFilterTokens': str(show_filter_tokens).lower(),
            },
            json=body,
        )
